from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from astropy import units as u
from astropy.constants import c as speed_of_light
from astropy.coordinates import EarthLocation
from astropy.io import fits
from spinifex.image_tools import (
    get_integrated_rm_from_fits,
    get_metadata_from_fits,
    get_rm_from_fits,
)
from spinifex.image_tools.fits_tools import get_freq_from_fits


def _make_fits_cube(
    extra_cards: dict[str, str] | None = None,
) -> fits.HDUList:
    header = fits.Header()
    header["NAXIS"] = 4
    header["NAXIS1"] = 10
    header["NAXIS2"] = 10
    header["NAXIS3"] = 6
    header["NAXIS4"] = 1
    header["CTYPE1"] = "RA---SIN"
    header["CRVAL1"] = 0
    header["CRPIX1"] = 5
    header["CDELT1"] = -(1 / 3600)
    header["CUNIT1"] = "deg"
    header["CTYPE2"] = "DEC--SIN"
    header["CRVAL2"] = 0
    header["CRPIX2"] = 5
    header["CDELT2"] = 1 / 3600
    header["CUNIT2"] = "deg"
    header["CTYPE3"] = "FREQ"
    header["CRVAL3"] = 1.4e9
    header["CRPIX3"] = 1
    header["CDELT3"] = 1e6
    header["CUNIT3"] = "Hz"
    header["CTYPE4"] = "STOKES"
    header["CRVAL4"] = 1
    header["CRPIX4"] = 1
    header["CDELT4"] = 1
    header["CUNIT4"] = ""
    header["BUNIT"] = "Jy/beam"
    header["DATE-OBS"] = (
        "2019-04-25T12:45:52.893302"  # pulled from a random SPICE-RACS file
    )

    if extra_cards is not None:
        for key, value in extra_cards.items():
            header[key] = value

    data = np.ones((1, 6, 10, 10), dtype=np.float32)

    hdu = fits.PrimaryHDU(data, header)

    return fits.HDUList([hdu])


@pytest.fixture
def simple_fits_cube(tmpdir):
    """Create a simple FITS cube"""

    simple_hdul = _make_fits_cube(
        {
            "DURATION": "1800",
            "TELESCOP": "ASKAP",
        }
    )

    fits_path = Path(tmpdir) / "simple.fits"
    simple_hdul.writeto(fits_path, output_verify="silentfix", overwrite=True)

    yield fits_path

    fits_path.unlink()


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
def test_get_freq_from_fits(simple_fits_cube):
    freqs = get_freq_from_fits(fits_path=simple_fits_cube)
    freqs_mhz = freqs.to("MHz").value
    test_vals = np.arange(1400, 1400 + 6, 1, dtype=np.float32)
    assert np.array_equal(freqs_mhz, test_vals)


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
def test_get_metadata_from_fits(simple_fits_cube):
    simple_metadata = get_metadata_from_fits(fits_path=simple_fits_cube)
    assert simple_metadata.duration.sec == 1800
    assert simple_metadata.start_time.fits == "2019-04-25T12:45:52.893"
    assert simple_metadata.name == "simple.fits"
    assert simple_metadata.location == EarthLocation.of_site("ASKAP")
    assert np.isclose(
        simple_metadata.source.ra.wrap_at(180 * u.deg).deg,
        0,
        atol=1 / 3600 * 2,
    )
    assert np.isclose(
        simple_metadata.source.dec.deg,
        0,
        atol=1 / 3600 * 2,
    )


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
def test_get_rm_from_fits(simple_fits_cube, tmpdir):
    rm = get_rm_from_fits(
        fits_path=simple_fits_cube,
        output_directory=Path(tmpdir),
    )
    expected_rm = np.array([-0.69612298, -0.6634822])
    assert np.allclose(rm.rm, expected_rm)


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
def test_get_integrated_rm_from_fits(simple_fits_cube, tmpdir):
    integrated_rm = get_integrated_rm_from_fits(
        fits_path=simple_fits_cube,
        output_directory=Path(tmpdir),
    )
    rm = get_rm_from_fits(
        fits_path=simple_fits_cube,
        output_directory=Path(tmpdir),
    )
    delta_angles = np.rad2deg(np.angle(integrated_rm.theta))

    freqs = get_freq_from_fits(simple_fits_cube)
    lsq = (speed_of_light / freqs).to("m") ** 2

    # Compute basic angle change from sum of angle change per timestep
    basic_angles = (
        (rm.rm[..., np.newaxis] * u.rad / u.m**2 * lsq).to("deg").sum(axis=0).value
    )

    assert np.allclose(delta_angles, basic_angles, atol=1e-1)



"""Testing of tomion ionospheric model"""

from __future__ import annotations

from importlib import resources

from astropy.utils import iers
from spinifex.ionospheric.models import parse_iono_kwargs
from spinifex.ionospheric.tomion_parser import TOMOION_FORMAT_DICT

iers.conf.auto_download = False

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from spinifex.geometry.get_ipp import IPP, get_ipp_from_skycoord
from spinifex.ionospheric import ionospheric_models


@pytest.fixture
def ipp() -> IPP:
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2023-01-08T01:00:00") + np.arange(0, 10) * 15 * u.min
    return get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )


@pytest.fixture
def ipp2() -> IPP:
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2023-01-08T01:00:00") + np.arange(0, 10) * 3.5 * u.hr
    return get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )


def test_ionosphere_tomion(ipp):
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        options = parse_iono_kwargs(
            ionospheric_models.tomion,
            output_directory=datapath,
        )
        tec = ionospheric_models.tomion(ipp, options=options)
        assert tec.electron_density.shape == ipp.loc.shape
        assert tec.electron_density_error.shape == ipp.loc.shape

        # Test bad arguments
        with pytest.raises(TypeError):
            options = parse_iono_kwargs(ionospheric_models.tomion, bad_arg="bad")


def test_ionosphere_tomionmultiple_days(ipp2):
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        options = parse_iono_kwargs(
            ionospheric_models.tomion,
            output_directory=datapath,
        )
        tec = ionospheric_models.tomion(ipp2, options=options)
        assert tec.electron_density.shape == ipp2.loc.shape
        assert tec.electron_density_error.shape == ipp2.loc.shape


def test_constants():
    # Previous constants from first implementation
    tomion_format = (
        "mjd index value stddev type number_of_observations height ra dec i j k \
        label longitude lst year doy month dom".split()
    )
    data_types = [
        float,
        int,
        float,
        float,
        str,
        int,
        float,
        float,
        float,
        int,
        int,
        int,
        str,
        float,
        float,
        int,
        int,
        int,
        int,
    ]

    assert list(TOMOION_FORMAT_DICT.keys()) == tomion_format
    assert list(TOMOION_FORMAT_DICT.values()) == data_types



"""Module to parse the IONosphere map EXchange (IONEX) data format,
as described in Schaer and Gurtner (1998)"""

from __future__ import annotations

import gzip
import io
from pathlib import Path
from typing import Any, NamedTuple, TextIO

import astropy.units as u
import numpy as np
from astropy.time import Time
from numpy.typing import NDArray
from unlzw3 import unlzw

from spinifex.exceptions import IonexError
from spinifex.ionospheric.ionex_download import IonexOptions
from spinifex.times import get_unique_days


class IonexData(NamedTuple):
    """Object containing all necessary information from Ionex data"""

    lons: NDArray[np.float64]
    """array with available longitude values (degrees)"""
    lats: NDArray[np.float64]
    """array with available latitude values (degrees)"""
    times: Time
    """available times"""
    dims: int
    """dimension of the heights (usually 1)"""
    h: NDArray[np.float64]
    """available heights (km)"""
    tec: NDArray[np.float64]
    """array with tecvalues times x lons x lats (TECU)"""
    rms: NDArray[np.float64]
    """array with rms of tecvalues times x lons x lats (TECU, if available, nan otherwise)"""


class IonexHeader(NamedTuple):
    """Object containing header information from ionex file"""

    lons: NDArray[np.float64]
    """array with available longitude values (degrees)"""
    lats: NDArray[np.float64]
    """array with available latitude values (degrees)"""
    times: Time
    """available times"""
    dims: int
    """dimension of the heights (usually 1)"""
    h: NDArray[np.float64]
    """available heights (km)"""
    mfactor: float
    """multiplication factor for tec values"""


def read_ionex(ionex_filename: Path, options: IonexOptions | None = None) -> IonexData:
    """Read and parse a ionex file. Returns a ionex object.

    Parameters
    ----------
    ionex_filename : str
        name of the ionex data file
    options : IonexOptions
        Options for the ionex model.
    Returns
    -------
    IonexData
        ionex object with data and grid

    """
    if ionex_filename.suffix == ".gz":
        with gzip.open(ionex_filename, "rt", encoding="utf-8") as file_buffer:
            return _read_ionex_data(file_buffer, options=options)

    if ionex_filename.suffix == ".Z":
        data = unlzw(ionex_filename.read_bytes()).decode("utf-8")
        with io.StringIO(data) as file_buffer:
            return _read_ionex_data(file_buffer, options=options)

    with ionex_filename.open(encoding="utf-8") as file_buffer:
        return _read_ionex_data(file_buffer, options=options)


def _read_ionex_header(filep: TextIO) -> IonexHeader:
    """Read header from ionex file. Put filepointer to the end of the header."""
    # Declare variables with types
    h1: float | None = None
    h2: float | None = None
    hstep: float | None = None
    start_lon: float | None = None
    end_lon: float | None = None
    step_lon: float | None = None
    start_lat: float | None = None
    end_lat: float | None = None
    step_lat: float | None = None
    start_time: Time | None = None
    ntimes: int | None = None
    step_time: u.Quantity | None = None
    mfactor: float | None = None
    dimension: int | None = None

    filep.seek(0)
    for line in filep:
        if "END OF HEADER" in line:
            break
        label = line[60:-1]
        record = line[:60]
        if "EPOCH OF FIRST MAP" in label:
            yy, mm, day, hr, minute, second = (
                int(float(i)) for i in record.strip().split()
            )
            epoch = Time(f"{yy}-{mm}-{day}T{hr % 24}:{minute}:{second}")
            start_time = epoch
        if "INTERVAL" in label:
            step_time = float(record) * u.s
        if "EXPONENT" in label:
            mfactor = 10.0 ** float(record)
        if "MAP DIMENSION" in label:
            dimension = int(record)
        if "HGT1 / HGT2 / DHGT" in label:
            h1, h2, hstep = (float(i) for i in record.split())
        if "LON1 / LON2 / DLON" in label:
            start_lon, end_lon, step_lon = (float(i) for i in record.split())
        if "LAT1 / LAT2 / DLAT" in label:
            start_lat, end_lat, step_lat = (float(i) for i in record.split())
        if "# OF MAPS IN FILE" in label:
            ntimes = int(record)

    # Check that all optional values are not None
    # Need to do this one by one to get MyPy to understand that they are not None
    # Should probably replace with higher level checks of the header values
    bad_msg = "Could not parse ionex header"
    if h1 is None:
        raise IonexError(bad_msg)
    if h2 is None:
        raise IonexError(bad_msg)
    if hstep is None:
        raise IonexError(bad_msg)
    if start_lon is None:
        raise IonexError(bad_msg)
    if end_lon is None:
        raise IonexError(bad_msg)
    if step_lon is None:
        raise IonexError(bad_msg)
    if start_lat is None:
        raise IonexError(bad_msg)
    if end_lat is None:
        raise IonexError(bad_msg)
    if step_lat is None:
        raise IonexError(bad_msg)
    if start_time is None:
        raise IonexError(bad_msg)
    if ntimes is None:
        raise IonexError(bad_msg)
    if step_time is None:
        raise IonexError(bad_msg)
    if mfactor is None:
        raise IonexError(bad_msg)
    if dimension is None:
        raise IonexError(bad_msg)

    harray = np.arange(h1, h2 + 0.5 * hstep, hstep) if hstep > 0 else np.array([h1])

    lonarray = np.arange(start_lon, end_lon + 0.5 * step_lon, step_lon)
    latarray = np.arange(start_lat, end_lat + 0.5 * step_lat, step_lat)
    timearray = start_time + np.arange(0, ntimes) * step_time

    return IonexHeader(
        mfactor=mfactor,
        lons=lonarray,
        lats=latarray,
        times=timearray,
        dims=dimension,
        h=harray,
    )


def _fill_data_record(
    data: NDArray[np.float64],
    filep: TextIO,
    stop_label: str,
    timeidx: int,
    ionex_header: IonexHeader,
) -> None:
    """Helper function to parse a data block of a single map in ionex.
    Puts filepointer to the end of the map

    Parameters
    ----------
    data : NDArray
        pre allocated array to store the datablock
    filep : TextIO
        _description_
    stop_label : str
        end of the data block indicator
    timeidx : int
        index of time of the data block
    ionex_header : namedtuple
        header information
    """
    line = filep.readline()  # read EPOCH (not needed since we have the index)
    # TODO: Properly type tec
    tec: list[Any] = []
    lonidx = 0
    latidx = 0
    # TODO: Replace magic numbers with named constants
    for line in filep:
        label = line[60:-1]
        if stop_label in label:
            if tec:
                tec_array = np.array(tec) * ionex_header.mfactor
                data[timeidx, lonidx:, latidx] = tec_array
            return
        if "LAT/LON1/LON2/DLON/H" in label:
            if tec:
                tec_array = np.array(tec) * ionex_header.mfactor
                data[timeidx, lonidx:, latidx] = tec_array
            tec = []
            record = line[:60]
            lat, lon1, _, _, _ = (float(record[i : i + 6]) for i in range(2, 32, 6))
            latidx = int(np.argmin(np.abs(ionex_header.lats - lat)))
            lonidx = int(np.argmin(np.abs(ionex_header.lons - lon1)))
        else:
            record = line[:-1]
            tec += [
                float(record[i : i + 5])
                for i in range(0, len(record), 5)
                if record[i : i + 5].strip()
            ]


def _read_ionex_data(filep: TextIO, options: IonexOptions | None = None) -> IonexData:
    """This function parses the IONEX file.
    Some fixed structure (like data records being strings of exactly 80 characters) of the file
    is assumed. This structure is described in Schaer and Gurtner (1998).

    Parameters
    ----------
    filep : TextIO
        pointer to an ionex file
    options : IonexOptions | None, optional
        options for ionex model, by default None

    Returns
    -------
    IonexData
        ionex object
    """
    if options is None:
        options = IonexOptions()
    ionex_header = _read_ionex_header(filep)
    tecarray = np.zeros(
        ionex_header.times.shape + ionex_header.lons.shape + ionex_header.lats.shape,
        dtype=float,
    )
    rmsarray = np.full(tecarray.shape, np.nan)
    for line in filep:
        # _read_ionex_header should have put the filep at the end of the header
        label = line[60:-1]
        record = line[:60]
        if "START OF TEC MAP" in label:
            timeidx = int(record) - 1
            _fill_data_record(tecarray, filep, "END OF TEC MAP", timeidx, ionex_header)
        if "START OF RMS MAP" in label:
            timeidx = int(record) - 1
            _fill_data_record(rmsarray, filep, "END OF RMS MAP", timeidx, ionex_header)
    if options.prefix == "uqr" and options.correct_uqrg_rms:
        # apply uqr correction zhao et al. (2021)
        rmsarray = rmsarray - 6
        rmsarray[rmsarray < 1] = 1

    return IonexData(
        lons=ionex_header.lons,
        lats=ionex_header.lats,
        times=ionex_header.times,
        dims=ionex_header.dims,
        h=ionex_header.h,
        tec=tecarray,
        rms=rmsarray,
    )


def unique_days_from_ionex(ionex_data: IonexData | list[IonexData]) -> Time:
    """Get unique days from a ionex object or list of ionex objects.

    Parameters
    ----------
    ionex_data : IonexData | list[IonexData]
        ionex object or list of ionex objects

    Returns
    -------
    Time
        unique days
    """
    # Get first MJD of each ionex object
    # This avoids issues with midnight crossing
    if isinstance(ionex_data, IonexData):
        time_jd_array = ionex_data.times.sort().mjd[0]
    else:
        time_list: list[NDArray[np.int64]] = []
        for ionex in ionex_data:
            ionex_time = ionex.times.sort().mjd[0]
            time_list.append(ionex_time)
        time_jd_array = np.array(time_list)

    times = Time(
        time_jd_array,
        format="mjd",
    )
    return get_unique_days(times)


def unique_days_from_ionex_files(ionex_files: list[Path] | Path) -> Time:
    """Get unique days from a list of ionex files.

    Parameters
    ----------
    ionex_files : list[Path]
        list of ionex files

    Returns
    -------
    Time
        unique days
    """

    if isinstance(ionex_files, Path):
        ionex_data = read_ionex(ionex_files)
        return unique_days_from_ionex(ionex_data)

    ionex_data_list = [read_ionex(ionex_file) for ionex_file in ionex_files]
    return unique_days_from_ionex(ionex_data_list)



"""Module for getting the Earth magnetic field"""

from __future__ import annotations

from spinifex.geometry.get_ipp import IPP, get_ipp_from_altaz, get_ipp_from_skycoord

__all__ = ["IPP", "get_ipp_from_altaz", "get_ipp_from_skycoord"]



"""Predict / correct ionospheric RM in images"""

# Follows report by Van Eck (2021) and
# Python implementation in FRion https://github.com/CIRADA-Tools/FRion/
# FRion License: MIT Copyright (c) 2021 Cameron Van Eck

from __future__ import annotations

from typing import NamedTuple

import astropy.units as u
import numpy as np
from astropy.constants import c as speed_of_light
from astropy.time import Time
from numpy.typing import NDArray
from scipy import integrate

from spinifex.get_rm import RM, get_average_rm
from spinifex.logger import logger


class IntegratedRM(NamedTuple):
    """Time-integrated rotation  measure"""

    theta: NDArray[np.complex128]
    """Complex time-integrated effect of the ionosphere"""
    time: Time
    """Average time"""
    b_parallel: float
    """Average parallel magnetic field"""
    electron_density: float
    """Average electron content"""
    height: float
    """Average altitude (km)"""
    azimuth: float
    """Average azimuth (degrees)"""
    elevation: float
    """Average elevation (degrees)"""


def get_integrated_rm(
    time_dep_rm: RM,
    freq_arr: u.Quantity,
) -> IntegratedRM:
    """Computed the integrated RM effect following Van Eck (2021)

    Parameters
    ----------
    time_dep_rm : RM
        Time-dependent RM object
    freq_arr : u.Quantity
        Frequency array

    Returns
    -------
    IntegratedRM
        Integrated rotation measure object
    """

    # Get required values for integration
    duration_seconds = (
        ((time_dep_rm.times.max() - time_dep_rm.times.min()).jd * u.day)
        .to(u.second)
        .value
    )
    times_mjd_seconds = (time_dep_rm.times.mjd * u.day).to(u.second).value

    wave_sq = (u.Quantity(speed_of_light) / freq_arr) ** 2
    rm_arr = time_dep_rm.rm * u.rad / u.m**2

    # Following Van Eck (2021) Eqn. 3 - compute integrated RM
    integrand = np.exp(
        2j * (wave_sq[:, np.newaxis] * rm_arr[np.newaxis]).to(u.rad).value
    )
    theta = (1 / (duration_seconds)) * integrate.simpson(
        y=integrand, x=times_mjd_seconds, axis=1
    )

    assert theta.shape == freq_arr.shape, (
        f"Wrong shape for theta ({theta.shape}), expected shape of {freq_arr.shape}"
    )

    # Following FRion implementation (https://github.com/CIRADA-Tools/FRion/)
    # report where numerical issues are likely to cause problems

    longest_wave_sq = np.max(wave_sq).to(u.m**2)
    largest_delta_rm = np.max(np.abs(np.diff(rm_arr)))
    largest_delta_pa = largest_delta_rm * longest_wave_sq

    if largest_delta_pa > 10 * u.deg:
        # Report warnings if the largest change in RM is greater than 10 degrees
        # This a slightly arbitrary threshold - FRion uses 0.5 rad ~ 28.6 degrees
        logger.warning(
            "Largest change of RM in time is greater than 10 degrees, "
            "this may cause numerical issues."
            "Consider using a smaller timestep."
        )

    for bad_fraction in (
        # Report warnings if the depolarisation is greater than 2% or 10%
        0.02,
        0.1,
    ):
        if np.min(np.abs(theta)) < bad_fraction:
            percentage = (1 - bad_fraction) * 100
            msg = (
                f"Depolarsation greater than {percentage:0.1f}% predicted!\n"
                "Images corrected by these values will likely be unreliable."
            )
            logger.warning(msg)
    averaged_rm = get_average_rm(time_dep_rm)
    return IntegratedRM(
        theta=theta,
        time=averaged_rm.times,
        b_parallel=averaged_rm.b_parallel,
        electron_density=averaged_rm.electron_density,
        height=averaged_rm.height,
        azimuth=averaged_rm.azimuth,
        elevation=averaged_rm.elevation,
    )



"""Module for getting the Earth magnetic field"""

from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple, Protocol

import astropy.units as u
import numpy as np
from astropy.coordinates import ITRS, AltAz
from ppigrf import igrf

from spinifex.geometry.get_ipp import IPP
from spinifex.times import get_unique_days


class MagneticProfile(NamedTuple):
    """Data object to hold Magnetic field profile and uncertainties"""

    magnetic_field: u.Quantity
    magnetic_field_error: u.Quantity


class MagneticFieldFunction(Protocol):
    """Magnetic field callable"""

    def __call__(self, ipp: IPP) -> MagneticProfile: ...


@dataclass
class MagneticModels:
    """Supported magnetic field models"""

    ppigrf: MagneticFieldFunction


def get_ppigrf_magnetic_field(ipp: IPP) -> u.Quantity:
    """Get the magnetic field at a given EarthLocation"""

    RMS_E = 87
    RMS_N = 73
    RMS_U = 114

    # constants from https://geomag.bgs.ac.uk/research/modelling/IGRF.html

    unique_days = get_unique_days(ipp.times)
    b_par = np.zeros(ipp.loc.shape, dtype=float)
    relative_uncertainty = np.zeros_like(b_par)
    for u_day in unique_days:
        indices = np.floor(ipp.times.mjd) == np.floor(u_day.mjd)
        loc = ipp.loc[indices]
        b_e, b_n, b_u = igrf(
            lon=loc.lon.deg,
            lat=loc.lat.deg,
            h=loc.height.to(u.km).value,
            date=u_day.to_datetime(),
        )
        # ppigrf adds an extra axis for time, we remove it by taking the first element
        b_magn = np.sqrt(b_e**2 + b_n**2 + b_u**2)[0]
        # relative uncertainty is 1/2 of the relative uncertainty (rms / b_magn**2) of the
        # sum of individual uncertainties of the squares (2 * rms_<enu> * b_<enu>)
        # multiply by b_magn to get absolute value
        rms = (RMS_E * b_e + RMS_N * b_n + RMS_U * b_u) / (b_magn)
        relative_uncertainty[indices] = rms / b_magn
        b_az = np.arctan2(b_e, b_n)
        b_el = np.arctan2(b_u, np.sqrt(b_n**2 + b_e**2))
        b_altaz = AltAz(az=b_az[0] * u.rad, alt=b_el[0] * u.rad, location=loc)
        b_itrs = b_altaz.transform_to(ITRS())

        # project to LOS
        los = ipp.los[indices][:, np.newaxis]
        b_par[indices] = los.x * b_itrs.x + los.y * b_itrs.y + los.z * b_itrs.z
        b_par[indices] *= b_magn
        # magnitude along LOS,

    return MagneticProfile(
        magnetic_field=u.Quantity(b_par * u.nanotesla),
        magnetic_field_error=u.Quantity(
            np.abs(b_par * relative_uncertainty) * u.nanotesla
        ),
    )


magnetic_models = MagneticModels(ppigrf=get_ppigrf_magnetic_field)


def parse_magnetic_model(magnetic_model_name: str) -> MagneticFieldFunction:
    """parse magnetic model name

    Parameters
    ----------
    magnetic_model_name : str
        name of the magnetic model

    Returns
    -------
    MagneticFieldFunction
        magnetic field function

    Raises
    ------
    TypeError
        if the magnetic model is not known

    """

    try:
        return getattr(magnetic_models, magnetic_model_name)  # type: ignore[no-any-return]
    except AttributeError as e:
        msg = f"Unknown magnetic model {magnetic_model_name}. Supported models are {list(magnetic_models.__annotations__.keys())}"
        raise TypeError(msg) from e



"""Spinifex tools for correcting image data"""

from __future__ import annotations

from .fits_tools import (
    FITSMetaData,
    get_freq_from_fits,
    get_integrated_rm_from_fits,
    get_metadata_from_fits,
    get_rm_from_fits,
)
from .image_tools import (
    IntegratedRM,
    get_integrated_rm,
)

__all__ = [
    "FITSMetaData",
    "IntegratedRM",
    "get_freq_from_fits",
    "get_integrated_rm",
    "get_integrated_rm_from_fits",
    "get_metadata_from_fits",
    "get_rm_from_fits",
]



"""Utilities for downloading IONEX files from various servers."""

from __future__ import annotations

import asyncio
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Literal

import astropy.units as u
from astropy.time import Time
from pydantic import Field

from spinifex.asyncio_wrapper import sync_wrapper
from spinifex.download import download_or_copy_url
from spinifex.exceptions import IonexError, TimeResolutionError
from spinifex.logger import logger
from spinifex.options import Options
from spinifex.times import get_gps_week, get_unique_days

# We need to support downloading from the following sources:
# "cddis.nasa.gov": cddis_nasa_gov,
# "chapman.upc.es": chapman_upc_es
# "igsiono.uwm.edu.pl": igsiono_uwm_edu_pl

CENTER_NAMES = {
    "cod",
    "esa",
    "igs",
    "jpl",
    "upc",
    "irt",
    "uqr",
}
DEFAULT_TIME_RESOLUTIONS: dict[str, u.Quantity] = {
    "cod": 1 * u.hour,
    "esa": 2 * u.hour,
    "igs": 2 * u.hour,
    "jpl": 2 * u.hour,
    "upc": 2 * u.hour,
    "irt": 2 * u.hour,
    "uqr": 15 * u.min,  # Chapman only provides 15 minute resolution right now
}


class Servers(str, Enum):
    CDDIS = ("cddis", "https://cddis.nasa.gov/archive/gnss/products/ionex")
    CHAPMAN = ("chapman", "http://chapman.upc.es/tomion/rapid")
    IGSIONO = ("igsiono", "ftp://igs-final.man.olsztyn.pl")

    def __new__(cls, value: str, _url: str) -> Servers:
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj._url = _url  # type: ignore[attr-defined]
        return obj

    @property
    def url(self) -> str:
        return str(self._url)  # type: ignore[attr-defined]

    @classmethod
    def get_url(cls, server: Servers) -> str:
        return str(server.url)  # Retrieve URL given an Enum value


assert set(DEFAULT_TIME_RESOLUTIONS.keys()) == CENTER_NAMES, (
    "Time resolutions must be defined for all analysis centres"
)
NAME_SWITCH_WEEK = 2238  # GPS Week where the naming convention changed

SOLUTION = Literal["final", "rapid"]


def new_cddis_format(
    time: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Get the URL for a new IONEX file from CDDIS.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Analysis centre prefix, by default "cod". Must be a supported analysis centre.
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".

    Returns
    -------
    str
        File URL
    """
    # Name Format Since GPS Week 2238
    # WWWW/IGS0OPSTYP_YYYYDDDHHMM_01D_SMP_CNT.INX.gz
    # Code	Meaning
    # WWWW	GPS week
    # TYP	solution type identifier -FIN (Final solution combination)
    # RAP (Rapid solution combination
    # YYYY	4-digit year
    # DDD	3-digit day of year
    # HH	2-digit hour
    # MM	2-digit minute
    # SMP	temporal product sampling resolution
    # CNT	content type -GIM (global ionosphere (TEC) maps)
    # ROT (rate of TEC index maps)
    # .gz	gzip compressed file
    assert time.isscalar, "Only one time is supported"
    if prefix not in CENTER_NAMES:
        msg = f"prefix {prefix} is not supported. Supported prefixes are {CENTER_NAMES}"
        raise IonexError(msg)

    # Parse time resolution
    if time_resolution is None:
        time_resolution = DEFAULT_TIME_RESOLUTIONS.get(prefix)
        if time_resolution is None:
            msg = f"Time resolution not defined for {prefix}"
            raise TimeResolutionError(msg)
        msg = f"Using default time resolution {time_resolution} for {prefix}"
        logger.info(msg)
    if not time_resolution.value.is_integer():
        error = f"Time resolution must be an integer. Got {time_resolution}"
        raise TimeResolutionError(error)
    if time_resolution.to(u.min).value % 15 != 0:
        msg = f"Time resolution on CDDIS is multiples 15 minutes. Please check the time resolution ({time_resolution})."
        logger.warning(msg)
    if time_resolution < 1 * u.hour:
        time_resolution = time_resolution.to(u.min)
    else:
        time_resolution = time_resolution.to(u.hour)
    time_res_str = (
        f"{int(time_resolution.value):02d}{str(time_resolution.unit)[0].upper()}"
    )

    if solution == "final":
        solution_str = "FIN"
    elif solution == "rapid":
        solution_str = "RAP"
    else:
        msg = f"Solution {solution} is not supported. Supported solutions are ['final', 'rapid']"  # type: ignore[unreachable]
        raise IonexError(msg)

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    prefix_formatted = prefix.upper()

    # WWWW/IGS0OPSTYP_YYYYDDDHHMM_01D_SMP_CNT.INX.gz
    file_name = f"{prefix_formatted}0OPS{solution_str}_{dtime.year:03d}{doy:03d}0000_01D_{time_res_str}_GIM.INX.gz"
    directory = f"{dtime.year:04d}/{doy:03d}"

    if url_stem is None:
        url_stem = Servers.get_url(Servers.CDDIS)

    return f"{url_stem}/{directory}/{file_name}"


def old_cddis_format(
    time: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Get the URL for an old IONEX file from CDDIS.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Analysis centre prefix, by default "cod"
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".

    Returns
    -------
    str
        File URL

    Raises
    ------
    IonexError
        If the prefix is not a supported analysis centre.
    """
    # Name Format Before GPS Week 2237
    # YYYY/DDD/AAAgDDD#.YYi.Z - Vertical total electron content (TEC) maps
    # Code	Meaning
    # YYYY	4-digit year
    # DDD	3-digit day of year
    # AAA	Analysis center name
    # #	file number for the day, typically 0
    # YY	2-digit year
    # .Z	Unix compressed file
    if time_resolution is not None:
        msg = f"Time resolution is not used by the old CDDIS format (got `{time_resolution}`). Ignoring."
        logger.warning(msg)

    assert time.isscalar, "Only one time is supported"
    if prefix not in CENTER_NAMES:
        msg = f"prefix {prefix} is not supported. Supported prefixes are {CENTER_NAMES}"
        raise IonexError(msg)

    if solution == "final":
        prefix_str = prefix
    elif solution == "rapid":
        prefix_str = prefix[:-1] + "r"
    else:
        msg = f"Solution {solution} is not supported. Supported solutions are ['final', 'rapid']"  # type: ignore[unreachable]
        raise IonexError(msg)

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    # YYYY/DDD/AAAgDDD#.YYi.Z
    yy = f"{dtime.year:02d}"[-2:]
    file_name = f"{prefix_str}g{doy:03d}0.{yy}i.Z"
    directory_name = f"{dtime.year:04d}/{doy:03d}"
    if url_stem is None:
        url_stem = Servers.get_url(Servers.CDDIS)

    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_cddis(
    times: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from CDDIS.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        Analysis centre prefix, by default "cod"
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".
    output_directory : Path | None, optional
        Output directory path, by default None, will default to `ionex_files` in the current working directory.

    Returns
    -------
    list[Path]
        List of downloaded files
    """
    unique_days: Time = get_unique_days(times)

    coros = []
    for day in unique_days:
        if get_gps_week(day) >= NAME_SWITCH_WEEK:
            formatter = new_cddis_format
        else:
            formatter = old_cddis_format
        url = formatter(
            day,
            prefix=prefix,
            url_stem=url_stem,
            time_resolution=time_resolution,
            solution=solution,
        )
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


def chapman_format(
    time: Time,
    prefix: str = "uqr",
    url_stem: str | None = None,
) -> str:
    """Format a URL for an IRTG file from Chapman.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Ionex prefix, by default "irt"
    url_stem : str | None, optional
        URL stem, by default None

    Returns
    -------
    str
        File URL
    """

    assert time.isscalar, "Only one time is supported"

    dtime: datetime = time.to_datetime()
    doy = int(time.datetime.timetuple().tm_yday)
    yy = f"{dtime.year:02d}"[-2:]
    # e.g. uqrg3260.96i.Z
    file_name = f"{prefix}g{doy:03d}0.{yy}i.Z"
    # e.g. 1996/326_961121.15min/
    # Hard coded to 15 minute resolution - Chapman only provides 15 minute resolution right now
    # Could be updated to be more flexible in the future
    directory_name = (
        f"{dtime.year:04d}/{doy:03d}_{yy}{dtime.month:02d}{dtime.day:02d}.15min"
    )
    if url_stem is None:
        url_stem = Servers.get_url(Servers.CHAPMAN)

    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_chapman(
    times: Time,
    prefix: str = "uqr",
    url_stem: str | None = None,
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from Chapman.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        Ionex prefix, by default "irt"
    url_stem : str | None, optional
        URL stem, by default None
    output_directory : Path | None, optional
        Output directory path, by default None

    Returns
    -------
    list[Path]
        List of downloaded file paths

    Raises
    ------
    IonexError
        If the date is before 2021/024.
    """
    unique_days: Time = get_unique_days(times)
    for day in unique_days:
        # Chapman goes back to 326th day of 1996 == 21/11/1996
        _chapman_start = datetime(1996, 11, 21)
        if day.datetime < _chapman_start:
            msg = f"Date {day.datetime} is before {_chapman_start}. Chapman only goes back to this date."
            raise IonexError(msg)

    coros = []
    for day in unique_days:
        url = chapman_format(day, prefix=prefix, url_stem=url_stem)
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


def igsiono_format(
    time: Time,
    prefix: str = "igs",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Format a URL for an IGS IONO file from IGSIONO.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Ionex prefix, by default "igs"
    time_resolution : u.Quantity | None, optional
        Ionex resolution, by default None
    url_stem : str | None, optional
        URL stem, by default None
    solution : SOLUTION, optional
        Type of solution, by default "final"

    Returns
    -------
    str
        File URL
    """
    assert time.isscalar, "Only one time is supported"

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    yy = f"{dtime.year:02d}"[-2:]
    directory_name = f"pub/gps_data/GPS_IONO/cmpcmb/{yy}{doy:03d}"
    if url_stem is None:
        url_stem = Servers.get_url(Servers.IGSIONO)

    # File name matches CDDIS format
    if get_gps_week(time) >= NAME_SWITCH_WEEK:
        formatter = new_cddis_format
    else:
        formatter = old_cddis_format
    file_name = Path(
        formatter(
            time,
            prefix=prefix,
            url_stem=url_stem,
            time_resolution=time_resolution,
            solution=solution,
        )
    ).name
    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_igsiono(
    times: Time,
    prefix: str = "igs",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from IGSIONO.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        URL prefix, by default "igs"
    time_resolution : u.Quantity | None, optional
        Ionex resolution, by default None
    url_stem : str | None, optional
        URL stem, by default None
    solution : SOLUTION, optional
        Type of solution, by default "final"
    output_directory : Path | None, optional
        Output directory path, by default None

    Returns
    -------
    list[Path]
        List of downloaded file paths
    """
    unique_days: Time = get_unique_days(times)

    coros = []
    for day in unique_days:
        url = igsiono_format(
            day,
            prefix=prefix,
            time_resolution=time_resolution,
            url_stem=url_stem,
            solution=solution,
        )
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


class IonexOptions(Options):
    """Options for ionex model"""

    server: Servers = Field(
        Servers.CHAPMAN, description="Server to download ionex files from"
    )
    prefix: str = Field("uqr", description="Analysis centre prefix")
    url_stem: str | None = Field(None, description="URL stem")
    time_resolution: u.Quantity | None = Field(
        None, description="Time resolution for ionex files"
    )
    solution: SOLUTION = Field("final", description="Solution type")
    output_directory: Path | None = Field(
        None, description="Output directory for ionex files"
    )
    correct_uqrg_rms: bool = Field(
        True, description="Correct overestimated rms of uqr maps"
    )


async def _download_ionex_coro(
    times: Time,
    options: IonexOptions | None = None,
) -> list[Path]:
    """Download IONEX files from a server.

    Parameters
    ----------
    times : Time
        Times to download for.
    options : IonexOptions
        Options for the ionex model.

    Returns
    -------
    list[Path]
        List of downloaded files

    Raises
    ------
    IonexError
        If the server is not supported.
    NotImplementedError
        If the server is not implemented yet.
    """
    if options is None:
        options = IonexOptions()

    # TODO: Consider refactor to pass around the options object
    if options.server == Servers.CDDIS:
        return await download_from_cddis(
            times,
            prefix=options.prefix,
            url_stem=options.url_stem,
            time_resolution=options.time_resolution,
            solution=options.solution,
            output_directory=options.output_directory,
        )
    if options.server == Servers.CHAPMAN:
        if options.prefix != "uqr":
            msg = f"Chapman server primarily supports uqr prefix. You provided {options.prefix}, this may fail"
            logger.warning(msg)

        return await download_from_chapman(
            times,
            prefix=options.prefix,
            url_stem=options.url_stem,
            output_directory=options.output_directory,
        )

    # Must be IGSIONO
    assert options.server == Servers.IGSIONO, "Server not supported"
    if options.prefix != "igs":
        msg = f"IGS IONO server primarily supports igs prefix. You provided {options.prefix}, this may fail"
        logger.warning(msg)
    return await download_from_igsiono(
        times,
        prefix=options.prefix,
        time_resolution=options.time_resolution,
        url_stem=options.url_stem,
        solution=options.solution,
        output_directory=options.output_directory,
    )


async def download_ionex_coro(
    times: Time,
    server: str = "chapman",
    prefix: str = "uqr",
    url_stem: str | None = None,
    time_resolution: u.Quantity | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from a server.

    Parameters
    ----------
    times : Time
        Times to download for.
    server : str
        Server to download from, by default "cddis". Must be a supported server.
    prefix : str, optional
        Analysis centre prefix, by default "cod". Must be a supported analysis centre.
    url_stem : str | None, optional
        URL stem, by default None, will default to the server URL.
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".
    output_directory : Path | None, optional
        Output directory path, by default None, will default to `ionex_files` in the current working directory.

    Returns
    -------
    list[Path]
        List of downloaded files

    Raises
    ------
    IonexError
        If the server is not supported.
    NotImplementedError
        If the server is not implemented yet.
    """

    options = IonexOptions(
        server=server,
        prefix=prefix,
        url_stem=url_stem,
        time_resolution=time_resolution,
        solution=solution,
        output_directory=output_directory,
    )

    return await _download_ionex_coro(times, options)


# Create synchronous wrapper of download_ionex_coro
download_ionex = sync_wrapper(download_ionex_coro)
_download_ionex = sync_wrapper(_download_ionex_coro)



"""Testing of the main module"""

from __future__ import annotations

from importlib import resources

from astropy.utils import iers

iers.conf.auto_download = False

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from spinifex import get_rm


def test_get_rm():
    """Test that get_rm does not crash"""
    times = Time("2020-01-08T01:00:00") + np.arange(10) * 25 * u.min
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        rm = get_rm.get_rm_from_skycoord(
            loc=dwingeloo,
            times=times,
            source=cas_a,
            iono_model_name="ionex",
            output_directory=datapath,
            prefix="cod",
            server="cddis",
        )
        assert isinstance(rm.rm, np.ndarray)
        assert rm.rm.shape == times.shape
        assert np.isclose(rm.rm[0], 0.0687, 0.001)
    average_rm = get_rm.get_average_rm(rm)
    assert np.isscalar(average_rm.rm)



"""Testing of mstools"""

from __future__ import annotations

import shutil
from importlib import resources
from pathlib import Path

from astropy.utils import iers

iers.conf.auto_download = False

import importlib.util

import astropy.units as u
import pytest

if not importlib.util.find_spec("casacore"):
    pytest.skip(
        "Skipping tests of ms_tools because casacore is not available.",
        allow_module_level=True,
    )

from spinifex.vis_tools.ms_tools import (
    get_columns_from_ms,
    get_dtec_from_ms,
    get_rm_from_ms,
)


@pytest.fixture
def unzip_ms(tmpdir) -> Path:  # type: ignore[misc]
    with resources.as_file(resources.files("spinifex.data.tests")) as test_data:
        zipped_ms = test_data / "test.ms.zip"
    shutil.unpack_archive(zipped_ms, tmpdir)

    yield Path(tmpdir / "test.MS")

    shutil.rmtree(tmpdir / "test.MS")


def test_unzip_worked(unzip_ms: Path):
    # Check that the unzipped directory exists
    assert unzip_ms.exists()


def test_mstools(unzip_ms: Path) -> None:
    cols = get_columns_from_ms(unzip_ms)
    assert "ANTENNA1" in cols
    with resources.as_file(resources.files("spinifex.data.tests")) as test_data:
        rms = get_rm_from_ms(
            unzip_ms,
            output_directory=test_data,
            prefix="esa",
            server="cddis",
            use_stations=["CS002HBA0"],
            timestep=20 * u.s,
        )
        assert "CS002HBA0" in rms
        dtec = get_dtec_from_ms(
            unzip_ms,
            output_directory=test_data,
            prefix="esa",
            server="cddis",
            use_stations=["CS002HBA0"],
            timestep=20 * u.s,
        )
        assert "CS002HBA0" in dtec


def test_station_selection(unzip_ms: Path) -> None:
    with resources.as_file(resources.files("spinifex.data.tests")) as test_data:
        rms = get_rm_from_ms(
            unzip_ms,
            use_stations=["CS002HBA0"],
            timestep=20 * u.s,
            output_directory=test_data,
            prefix="esa",
            server="cddis",
        )
        assert "CS002HBA0" in rms
        assert len(rms) == 1

        rms = get_rm_from_ms(
            unzip_ms,
            use_stations="all",
            timestep=20 * u.s,
            output_directory=test_data,
            prefix="esa",
            server="cddis",
        )
        stations = ["CS002HBA0", "RS210HBA", "RS509HBA"]
        assert all(station in rms for station in stations)
        assert len(rms) == 3

        rms = get_rm_from_ms(
            unzip_ms,
            use_stations="average",
            timestep=20 * u.s,
            output_directory=test_data,
            prefix="esa",
            server="cddis",
        )
        assert "average_station_pos" in rms
        assert len(rms) == 1



from __future__ import annotations

from pathlib import Path
from typing import Any

import astropy.units as u
import h5py
import numpy as np
from numpy.typing import NDArray

from spinifex.get_dtec import DTEC
from spinifex.get_rm import RM


def get_minimal_solset_name(h5file: h5py.File) -> str:
    """Find the best name for your new solset in an h5parm file. Default name is sol###

    Parameters
    ----------
    h5file : h5py.File
        the h5parm file

    Returns
    -------
    str
        frst available name for the solset
    """
    for solset_idx in range(1000):
        if f"sol{solset_idx:03d}" not in h5file:
            return f"sol{solset_idx:03d}"
    return "sol9999"


def get_minimal_soltab_name(solset: h5py.Dataset, soltab_type: str) -> str:
    """Find the best name for your new soltab in an h5parm file. Default name is [soltab_type]###

    Parameters
    ----------
    solset : h5py.Dataset
        the solset in an h5parm
    soltab_type : str
        type of the solution table

    Returns
    -------
    str
         frst available name for the soltab
    """

    for soltab_idx in range(1000):
        if f"{soltab_type}{soltab_idx:03d}" not in solset:
            return f"{soltab_type}{soltab_idx:03d}"
    return f"{soltab_type}9999"


def create_solset(h5file: h5py.File, solset_name: str | None = None) -> h5py.Dataset:
    """Create new empty solset in your h5parm file. A solset has an antenna and source table. Default name is sol###"

    Parameters
    ----------
    h5file : h5py.File
        the h5parm file, must be writable
    solset_name : str, optional
        name of the solset, if None the first available from sol### will be used, by default None

    Returns
    -------
    h5py.Dataset
        pointer to the new solset
    """
    if solset_name is None:
        solset_name = get_minimal_solset_name(h5file)
    if solset_name in h5file:
        message = f"{solset_name} already exists in {h5file.name}."
        raise RuntimeError(message)

    solset = h5file.create_group(solset_name)
    solset.attrs.h5parm_version = "1.0"
    solset.attrs["CLASS"] = np.bytes_("GROUP")
    solset.attrs["FILTERS"] = 0
    solset.attrs["TITLE"] = np.bytes_("")
    solset.attrs["VERSION"] = np.bytes_("1.0")
    solset.create_dataset(
        "antenna",
        shape=(1,),
        maxshape=(None,),
        chunks=True,
        dtype=[("name", "S16"), ("position", "<f4", (3,))],
    )
    solset.create_dataset(
        "source",
        shape=(1,),
        maxshape=(None,),
        chunks=True,
        dtype=[("name", "S16"), ("dir", "<f4", (2,))],
    )
    return solset


def add_antenna_info(
    solset: h5py.Dataset, antenna_names: NDArray[Any], antenna_pos: NDArray[Any]
) -> None:
    """Add antenna metadata to a solset

    Parameters
    ----------
    solset : h5py.Dataset
        A solset in an h5parm
    antenna_names : NDArray[Any]
        array with antenna names
    antenna_pos : NDArray[Any]
        array ant x 3 with ITRF antenna positions
    """
    ant_meta = solset["antenna"]
    ant_meta.resize((len(antenna_names),))
    for antenna_number, (antenna_name, pos) in enumerate(
        zip(antenna_names, antenna_pos)
    ):
        ant_meta[antenna_number] = (antenna_name, pos)


def add_source_info(
    solset: h5py.Dataset, source_names: NDArray[Any], source_dirs: NDArray[Any]
) -> None:
    """Add source metadata to a solset

    Parameters
    ----------
    solset : h5py.Dataset
        A solset in an h5parm
    source_names : NDArray[Any]
        array with source names
    source_dirs : NDArray[Any]
        array source x 2 with Ra,DEC directions of the sources
    """
    source_meta = solset["source"]
    source_meta.resize((len(source_names),))
    for source_number, (source_name, source_dir) in enumerate(
        zip(source_names, source_dirs)
    ):
        source_meta[source_number] = (source_name, source_dir)


def add_soltab(
    solset: h5py.Dataset,
    soltab_type: str,
    val: NDArray[Any],
    weight: NDArray[Any],
    soltab_axes: list[Any],
    axes_values: dict[str, Any],
    soltab_name: str | None = None,
) -> h5py.Dataset:
    """Add asolution table to an existing solset

    Parameters
    ----------
    solset : h5py.Dataset
        solset
    soltab_type : str
        type of the solutions
    val : NDArray[Any]
        array with solutions values, must match the shape of the axes
    weight : NDArray[Any]
        array of weights, must match the shape of val
    soltab_axes : list
        ordered list of axes as they are defined for val/wegiht
    axes_values : dict[Any]
        dictionary with the values for each axis in soltab_axes
    soltab_name : str|None, optional
        name of the solution table, if None or existing will default to <soltab_type>###, by default None

    Returns
    -------
    h5py.Dataset
        the solution table
    """
    if soltab_name is None:
        soltab_name = get_minimal_soltab_name(solset, soltab_type)
    if soltab_name in solset:
        message = f"{soltab_name} already exists in {solset.name}."
        raise RuntimeError(message)

    soltab = solset.create_group(soltab_name)
    soltab.attrs["CLASS"] = np.bytes_("GROUP")
    soltab.attrs["FILTERS"] = 0
    soltab.attrs["TITLE"] = np.bytes_(soltab_type)
    soltab.attrs["VERSION"] = np.bytes_("1.0")
    soltab_axes_str = ",".join(soltab_axes)
    for axis in soltab_axes:
        soltab.create_dataset(axis, data=axes_values[axis])

    soltab_val = soltab.create_dataset("val", shape=val.shape, dtype="<f8")
    soltab_val.attrs["soltype"] = soltab_type
    soltab_axes_dtype = _zero_terminated_string(len(soltab_axes_str) + 1)
    soltab_val.attrs["AXES"] = np.array(soltab_axes_str, dtype=soltab_axes_dtype)
    soltab_val[:] = val
    soltab_weight = soltab.create_dataset("weight", shape=weight.shape, dtype="<f2")
    soltab_weight.attrs["AXES"] = np.array(soltab_axes_str, dtype=soltab_axes_dtype)
    soltab_weight[:] = weight  # Todo: np.isnan?
    return soltab


def check_h5parm(h5file: h5py.File) -> bool:
    """Check if given h5file is suitable as h5parm

    Parameters
    ----------
    h5file : h5py.File
        possible h5parm file

    Returns
    -------
    bool
        True if the file is conform h5parm definition
    """
    return "TITLE" in h5file.attrs


def _get_station_metadata(rms: dict[str, Any]) -> tuple[list[Any], list[Any]]:
    """helper function to generate arrays of station names and positions

    Parameters
    ----------
    rms : dict[RM|DTEC]
        dictionary with RM object per station

    Returns
    -------
    tuple
        station_names and station_positions (ITRF xyz)
    """
    station_names = sorted(rms.keys())
    station_pos = [
        [
            rms[st_name].loc.to(u.m).x.value,
            rms[st_name].loc.to(u.m).y.value,
            rms[st_name].loc.to(u.m).z.value,
        ]
        for st_name in station_names
    ]
    return station_names, station_pos


def create_empty_h5parm(h5parm_name: str) -> None:
    """create an empty h5parm file

    Parameters
    ----------
    h5parm_name : str
        name of the file

    Raises
    ------
    TypeError
       error if h5parm_name is existing and not a valid h5parm
    """
    if Path(h5parm_name).is_file():
        if not h5py.is_hdf5(h5parm_name):
            msg = f"{h5parm_name} exists and not a valid hdf5 file"
            raise ValueError(msg)

        with h5py.File(h5parm_name, "r") as h5parm:
            if not check_h5parm(h5parm):
                msg = f"{h5parm_name} exists and not a valid h5parm file"
                raise ValueError(msg)
    else:
        with h5py.File(h5parm_name, "w") as h5parm:
            h5parm.attrs["h5parm_version"] = np.bytes_("1.0")
            h5parm.attrs["CLASS"] = np.bytes_("GROUP")
            h5parm.attrs["FILTERS"] = 0
            h5parm.attrs["TITLE"] = np.bytes_("")
            h5parm.attrs["VERSION"] = np.bytes_("1.0")


def _zero_terminated_string(size: int = 10) -> h5py.Datatype:
    tid = h5py.h5t.C_S1.copy()
    tid.set_size(size)
    return h5py.Datatype(tid)


def write_rm_to_h5parm(
    rms: dict[str, RM],
    h5parm_name: str,
    solset_name: str | None = None,
    soltab_name: str | None = None,
    add_to_existing_solset: bool = False,
) -> None:
    """writes a dictionary of RM values per station to a new or existing h5parm file

    Parameters
    ----------
    rms : dict[RM]
        rm values per station
    h5parm_name : str
        name of the h5parm file
    solset_name : str | None, optional
        name of the solset if None it  will default to sol###, by default None
    soltab_name : str | None, optional
        name of the soltab if None it  will default to 'rotationmeasure###'
    add_to_existing_solset : bool = False
        whether to append to an existing solset, if it exists. If True, the user
        is responsible for having consistent antennas and sources.

    Raises
    ------
    TypeError
        error if h5parm_name is existing and not a valid h5parm
    """
    create_empty_h5parm(h5parm_name=h5parm_name)
    station_names, station_pos = _get_station_metadata(rms)
    with h5py.File(h5parm_name, "a") as h5parm:
        if solset_name is not None and solset_name in h5parm and add_to_existing_solset:
            solset = h5parm[solset_name]
        else:
            solset = create_solset(h5parm, solset_name=solset_name)
            add_antenna_info(solset, station_names, station_pos)

        soltab_axes = ["ant", "time"]
        axes_values = {}
        ant_dtype = _zero_terminated_string(max(map(len, station_names)) + 1)
        axes_values["ant"] = np.array(station_names, dtype=ant_dtype)
        axes_values["time"] = (
            rms[station_names[0]].times.mjd * 24 * 3600.0
        )  # mjd in seconds?
        rm_values = np.array([rms[stname].rm for stname in station_names])
        weights = np.ones(rm_values.shape, dtype=bool)
        weights[np.isnan(rm_values)] = 0
        add_soltab(
            solset=solset,
            soltab_type="rotationmeasure",
            val=rm_values,
            weight=weights,
            soltab_axes=soltab_axes,
            axes_values=axes_values,
            soltab_name=soltab_name,
        )


def write_tec_to_h5parm(
    dtec: dict[str, DTEC],
    h5parm_name: str,
    solset_name: str | None = None,
    soltab_name: str | None = None,
    add_to_existing_solset: bool = False,
) -> None:
    """writes a dictionary of RM values per station to a new or existing h5parm file

    Parameters
    ----------
    dtec : dict[DTEC]
        electron density profiles per station
    h5parm_name : str
        name of the h5parm file
    solset_name : str | None, optional
        name of the solset if None it  will default to sol###, by default None
    soltab_name : str | None, optional
        name of the soltab if None it  will default to 'rotationmeasure###'
    add_to_existing_solset : bool = False
        whether to append to an existing solset, if it exists. If True, the user
        is responsible for having consistent antennas and sources.

    Raises
    ------
    TypeError
        error if h5parm_name is existing and not a valid h5parm
    """
    create_empty_h5parm(h5parm_name=h5parm_name)
    station_names, station_pos = _get_station_metadata(dtec)
    with h5py.File(h5parm_name, "a") as h5parm:
        if solset_name is not None and solset_name in h5parm and add_to_existing_solset:
            solset = h5parm[solset_name]
        else:
            solset = create_solset(h5parm, solset_name=solset_name)
            add_antenna_info(solset, station_names, station_pos)

        soltab_axes = ["ant", "time"]
        axes_values = {}
        ant_dtype = _zero_terminated_string(max(map(len, station_names)) + 1)
        axes_values["ant"] = np.array(station_names, dtype=ant_dtype)
        axes_values["time"] = (
            dtec[station_names[0]].times.mjd * 24 * 3600.0
        )  # mjd in seconds?
        dtec_values = np.array(
            [
                np.sum(dtec[stname].electron_density * dtec[stname].airmass, axis=-1)
                for stname in station_names
            ]
        )
        weights = np.ones(dtec_values.shape, dtype=bool)
        weights[np.isnan(dtec_values)] = 0
        add_soltab(
            solset=solset,
            soltab_type="tec",
            val=dtec_values,
            weight=weights,
            soltab_axes=soltab_axes,
            axes_values=axes_values,
            soltab_name=soltab_name,
        )



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class Options(BaseModel):
    """Base model options class (immutable by default)."""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True, extra="forbid")



from __future__ import annotations

import filecmp
import shutil
from importlib import resources
from pathlib import Path

from astropy.utils import iers

iers.conf.auto_download = False

import astropy.units as u
import pytest
from astropy.time import Time
from spinifex.exceptions import IonexError, TimeResolutionError
from spinifex.ionospheric import ionex_download
from unlzw3 import unlzw

pytest_plugins = ("pytest_asyncio",)


@pytest.fixture
def times() -> Time:
    times_str = [
        "1999-02-01T00:00:00.00",
        "2010-02-01T00:00:00",
        "2024-02-01T00:00:00",
        "2024-02-01T01:00:00",
    ]
    return Time(times_str)


@pytest.fixture
def old_time() -> Time:
    times_str = [
        "1989-02-01T00:00:00.00",
    ]
    return Time(times_str)


@pytest.fixture
def new_time() -> Time:
    times_str = [
        "2025-01-01T00:00:00.00",
    ]
    return Time(times_str)


@pytest.fixture
def igsiono_time() -> Time:
    times_str = "2024-12-14T00:00:00.00"
    return Time(times_str)


def test_old_cddis_format(times):
    time = times[0]
    prefix = "cod"

    # time resolution is not used!
    for time_resolution in (None, 30 * u.min, 2 * u.hour):
        url = ionex_download.old_cddis_format(
            time, prefix=prefix, time_resolution=time_resolution
        )
        assert (
            url
            == "https://cddis.nasa.gov/archive/gnss/products/ionex/1999/032/codg0320.99i.Z"
        )
    prefix = "bad"
    with pytest.raises(IonexError):
        url = ionex_download.old_cddis_format(time, prefix=prefix)

    prefix = "esa"
    url = ionex_download.old_cddis_format(time, prefix=prefix)
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/1999/032/esag0320.99i.Z"
    )

    url_stem = "my_stem"
    url = ionex_download.old_cddis_format(time, url_stem=url_stem)
    assert url == "my_stem/1999/032/codg0320.99i.Z"

    url = ionex_download.old_cddis_format(time, prefix="cod", solution="rapid")
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/1999/032/corg0320.99i.Z"
    )


def test_new_cddis_format(times):
    time = times[-1]
    url = ionex_download.new_cddis_format(time)
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/2024/032/COD0OPSFIN_20240320000_01D_01H_GIM.INX.gz"
    )

    url_stem = "my_stem"
    url = ionex_download.new_cddis_format(time, url_stem=url_stem)
    assert url == "my_stem/2024/032/COD0OPSFIN_20240320000_01D_01H_GIM.INX.gz"

    time_resolution = 2 * u.hour
    url = ionex_download.new_cddis_format(time, time_resolution=time_resolution)
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/2024/032/COD0OPSFIN_20240320000_01D_02H_GIM.INX.gz"
    )

    time_resolution = 30 * u.min
    url = ionex_download.new_cddis_format(time, time_resolution=time_resolution)
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/2024/032/COD0OPSFIN_20240320000_01D_30M_GIM.INX.gz"
    )

    url = ionex_download.new_cddis_format(time, solution="rapid")
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/2024/032/COD0OPSRAP_20240320000_01D_01H_GIM.INX.gz"
    )

    url = ionex_download.new_cddis_format(time, prefix="esa")
    assert (
        url
        == "https://cddis.nasa.gov/archive/gnss/products/ionex/2024/032/ESA0OPSFIN_20240320000_01D_02H_GIM.INX.gz"
    )

    with pytest.raises(IonexError):
        url = ionex_download.new_cddis_format(time, prefix="bad")

    with pytest.raises(TimeResolutionError):
        url = ionex_download.new_cddis_format(time, time_resolution=1.5 * u.min)


def test_chapman_format(times):
    base_url = "http://chapman.upc.es/tomion/rapid"
    expected_urls = [
        "1999/032_990201.15min/uqrg0320.99i.Z",
        "2010/032_100201.15min/uqrg0320.10i.Z",
        "2024/032_240201.15min/uqrg0320.24i.Z",
        "2024/032_240201.15min/uqrg0320.24i.Z",
        "2025/001_250101.15min/uqrg0010.25i.Z",
    ]
    for time, expected_url in zip(times, expected_urls):
        url = ionex_download.chapman_format(time, url_stem=base_url)
        assert url == f"{base_url}/{expected_url}"


@pytest.mark.asyncio
async def test_chapman_download(tmpdir, old_time, new_time):
    with pytest.raises(IonexError):
        await ionex_download.download_from_chapman(
            times=old_time,
            url_stem=None,
            output_directory=Path(tmpdir),
        )

    downloaded_files = await ionex_download.download_from_chapman(
        times=new_time,
        url_stem=None,
        output_directory=Path(tmpdir),
    )
    downloaded_file = downloaded_files[0]

    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        expected_file_compressed = datapath / "uqrg0010.25i.truncated.zip"
        shutil.unpack_archive(expected_file_compressed, tmpdir)

    expected_file = Path(tmpdir) / "uqrg0010.25i.truncated"

    downloaded_data = (
        unlzw(
            downloaded_file.read_bytes(),
        )
        .decode("utf-8")
        .splitlines()
    )

    expected_data = expected_file.read_text().splitlines()
    n_lines_tructed = len(expected_data) - 1  # Last line is truncated

    # Compare first 49 lines of the downloaded file
    for i, (line1, line2) in enumerate(zip(downloaded_data, expected_data)):
        if i == n_lines_tructed:
            break
        assert line1 == line2

    # Clean up
    downloaded_file.unlink(missing_ok=True)
    expected_file.unlink(missing_ok=True)


def test_igsiono_format(igsiono_time):
    url = ionex_download.igsiono_format(igsiono_time)
    assert (
        url
        == "ftp://igs-final.man.olsztyn.pl/pub/gps_data/GPS_IONO/cmpcmb/24349/IGS0OPSFIN_20243490000_01D_02H_GIM.INX.gz"
    )


@pytest.mark.asyncio
async def test_igsiono_download(tmpdir, igsiono_time):
    downloaded_files = await ionex_download.download_from_igsiono(
        times=igsiono_time,
        output_directory=Path(tmpdir),
    )
    downloaded_file = downloaded_files[0]

    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        expected_file = datapath / "IGS0OPSFIN_20243490000_01D_02H_GIM.INX.gz"

    assert filecmp.cmp(downloaded_file, expected_file)

    # Clean up
    downloaded_file.unlink(missing_ok=True)


def test_download_ionex_igsiono(tmpdir, igsiono_time):
    downloaded_files = ionex_download.download_ionex(
        server="igsiono",
        times=igsiono_time,
        prefix="igs",
        url_stem=None,
        time_resolution=None,
        solution="final",
        output_directory=Path(tmpdir),
    )
    downloaded_file = downloaded_files[0]

    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        expected_file = datapath / "IGS0OPSFIN_20243490000_01D_02H_GIM.INX.gz"

    assert filecmp.cmp(downloaded_file, expected_file)

    # Clean up
    downloaded_file.unlink(missing_ok=True)


def test_download_ionex_chapman(tmpdir, new_time):
    downloaded_files = ionex_download.download_ionex(
        server="chapman",
        times=new_time,
        prefix="uqr",
        url_stem=None,
        time_resolution=None,
        solution="final",
        output_directory=Path(tmpdir),
    )
    downloaded_file = downloaded_files[0]

    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        expected_file_compressed = datapath / "uqrg0010.25i.truncated.zip"
        shutil.unpack_archive(expected_file_compressed, tmpdir)

    expected_file = Path(tmpdir) / "uqrg0010.25i.truncated"

    downloaded_data = (
        unlzw(
            downloaded_file.read_bytes(),
        )
        .decode("utf-8")
        .splitlines()
    )

    expected_data = expected_file.read_text().splitlines()
    n_lines_tructed = len(expected_data) - 1  # Last line is truncated

    # Compare first 49 lines of the downloaded file
    for i, (line1, line2) in enumerate(zip(downloaded_data, expected_data)):
        if i == n_lines_tructed:
            break
        assert line1 == line2

    # Clean up
    downloaded_file.unlink(missing_ok=True)
    expected_file.unlink(missing_ok=True)



#  Copyright (C) 2024 ASTRON (Netherlands Institute for Radio Astronomy) &
#  CSIRO (Commonwealth Scientific and Industrial Research Organisation)
#  SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pathlib import Path

from spinifex import __version__

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
    "myst_parser",
    "sphinx_autodoc_typehints",
    "autoapi.extension",
    "nbsphinx",
    "sphinxarg.ext",
]

autoapi_type = "python"
autoapi_dirs = ["../../spinifex"]
autoapi_member_order = "groupwise"
autoapi_keep_files = False
autoapi_root = "autoapi"
autoapi_add_toctree_entry = True

# Assumes tox is used to call sphinx-build
project_root_directory = Path.cwd().as_posix()
# The suffix of source filenames.
source_suffix = [".rst"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "spinifex"
copyright = "2025, ASTRON & CSIRO"

# openstackdocstheme options
repository_name = "git.astron.nl/spinifex"
bug_project = "none"
bug_tag = ""
html_last_updated_fmt = "%Y-%m-%d %H:%M"

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

version = __version__

modindex_common_prefix = ["spinifex."]

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
html_theme = "furo"
html_logo = "spinifex-logo.svg"
# html_theme = "sphinx_rtd_theme"
html_static_path = ["static"]
html_css_files = [
    "css/custom.css",
]

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}doc"

# Conf.py variables exported to sphinx rst files access using |NAME|
variables_to_export = [
    "project",
    "copyright",
    "version",
]

# Write to rst_epilog to export `variables_to_export` extract using `locals()`
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-rst_epilog
# frozen_locals = dict(locals())
# rst_epilog = "\n".join(
#     map(
#         lambda x: f".. |{x}| replace:: {frozen_locals[x]}",
#         variables_to_export,
#     )
# )
# # Pep is not able to determine that frozen_locals always exists so noqa
# del frozen_locals



"""Module to get iri density profile"""

from __future__ import annotations

import astropy.units as u
import numpy as np
from numpy.typing import NDArray
from PyIRI import coeff_dir
from PyIRI.main_library import IRI_density_1day

from spinifex.geometry.get_ipp import IPP
from spinifex.times import get_unique_days


def get_profile(ipp: IPP) -> NDArray[np.float32]:
    """Get the normalized electron density profile for all times an altitudes in ipp

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints

    Returns
    -------
    NDArray
        normalied density profile per time
    """
    unique_days = get_unique_days(times=ipp.times)
    edp = np.zeros(ipp.loc.shape, dtype=float)  # electron density profile
    altitudes = ipp.loc.height.to(u.km).value
    longitudes = ipp.loc.lon.deg
    latitudes = ipp.loc.lat.deg
    f107 = 100
    ccir_or_ursi = 0
    for u_day in unique_days:
        year, month, day, _, _, _ = u_day.ymdhms
        indices = np.floor(ipp.times.mjd) == np.floor(u_day.mjd)
        index_nr = np.where(indices)[0]
        aalt = altitudes[indices]  # iri input array of heights

        hidx = np.argmin(
            np.abs(aalt - 350), axis=1
        )  # use lon lat closest to altitude of 350 km for profile
        alon = longitudes[indices, hidx]  # iri input array of longitudes
        alat = latitudes[indices, hidx]  # iri input array of latitude
        ahr = ipp.times[indices].ymdhms.hour
        # IRI_density_1day treats lon/lat and hr as two separate arrays with independent lengths
        # I do not see another solution than to loop
        for itime, tmidx in zip(range(ahr.shape[0]), index_nr):
            _, _, _, _, _, _, edpi = IRI_density_1day(
                year,
                month,
                day,
                ahr[itime : itime + 1],
                alon[itime : itime + 1],
                alat[itime : itime + 1],
                aalt[itime],
                f107,
                coeff_dir,
                ccir_or_ursi,
            )
            edp[tmidx] = edpi.squeeze()
    edp /= np.sum(edp, axis=1, keepdims=True)
    return edp



"""Generic download utils"""

from __future__ import annotations

import asyncio
import shutil
from ftplib import FTP
from pathlib import Path
from urllib.parse import urlparse

import requests

from spinifex.exceptions import IonexError
from spinifex.logger import logger


def _ftp_download_and_quit(ftp: FTP, file_name: str, output_file: Path) -> None:
    """Download a file from an FTP server and quit the connection.

    Parameters
    ----------
    ftp : FTP
        FTP connection.
    file_name : str
        File name to download.
    output_file : Path
        Output file path.

    Raises
    ------
    e
        If the download fails.
    """
    try:
        with output_file.open("wb") as file_desc:
            ftp.retrbinary(f"RETR {file_name}", file_desc.write)
    except Exception as e:
        output_file.unlink(missing_ok=True)
        raise e
    finally:
        ftp.quit()


async def download_file_ftp(
    url: str,
    output_file: Path,
) -> None:
    """Download a file from a given URL using asyncio.

    Parameters
    ----------
    url : str
        URL to download.
    output_file : Path
        Output file path.
    """
    url_parsed = urlparse(url)
    url_path = Path(url_parsed.path)
    file_name = url_path.name
    directory_name = url_path.parent.as_posix()[1:]  # Remove leading slash

    ftp = FTP(url_parsed.netloc)
    # Anonymous login
    ftp.login()
    ftp.cwd(directory_name)

    await asyncio.to_thread(_ftp_download_and_quit, ftp, file_name, output_file)


async def download_file_http(
    url: str,
    output_file: Path,
    timeout_seconds: int = 30,
    chunk_size: int = 1000,
) -> None:
    """Download a file from a given URL using asyncio.

    Parameters
    ----------
    url : str
        URL to download.
    output_file : Path
        Output file path.
    timeout_seconds : int, optional
        Seconds to wait for request timeout, by default 30
    chunk_size : int, optional
        Chunks of data to download, by default 1000

    Raises
    ------
    IonexError
        If the download times out.
    """
    msg = f"Downloading from {url}"
    logger.info(msg)
    try:
        response = await asyncio.to_thread(requests.get, url, timeout=timeout_seconds)
    except requests.exceptions.Timeout as e:
        msg = "Timed out connecting to server"
        logger.error(msg)
        raise IonexError(msg) from e

    response.raise_for_status()
    with output_file.open("wb") as file_desc:
        for chunk in response.iter_content(chunk_size=chunk_size):
            await asyncio.to_thread(file_desc.write, chunk)


async def download_or_copy_url(
    url: str,
    output_directory: Path | None = None,
    chunk_size: int = 1000,
    timeout_seconds: int = 30,
) -> Path:
    """Download a file from a given URL.

    If the URL is a file URL (i.e. starting with `file://`), it will be copied to the output directory.

    Parameters
    ----------
    url : str
        URL to download.
    output_directory : Path | None, optional
        Output directory, by default None. If None, will default to `ionex_files` in the current working directory.
    chunk_size : int, optional
        Download chunks, by default 1000
    timeout_seconds : int, optional
        Request timeout in seconds, by default 30

    Returns
    -------
    Path
        Output file path

    Raises
    ------
    FileNotFoundError
        If the .netrc file is not found when downloading from CDDIS.
    """
    if output_directory is None:
        output_directory = Path.cwd() / "ionex_files"

    output_directory.mkdir(exist_ok=True)

    url_parsed = urlparse(url)
    url_path = Path(url_parsed.path)
    file_name = url_path.name
    output_file = output_directory / file_name

    if output_file.exists():
        msg = f"File {output_file} already exists. Skipping download."
        logger.info(msg)
        return output_file

    if url_parsed.scheme == "file":
        msg = f"URL scheme {url_parsed.scheme} is not supported"
        logger.info(msg)
        result = await asyncio.to_thread(shutil.copy, url_path, output_file)
        return Path(result)

    if url_parsed.scheme == "ftp":
        await download_file_ftp(url, output_file)
        return output_file

    if url.startswith("https://cddis.nasa.gov"):
        # CDDIS requires a .netrc file to download
        netrc = Path("~/.netrc").expanduser()
        if not netrc.exists():
            msg = "See: https://cddis.nasa.gov/Data_and_Derived_Products/CreateNetrcFile.html"
            logger.error(msg)
            msg = "Please add your NASA Earthdata login credentials to ~/.netrc"
            logger.error(msg)
            raise FileNotFoundError(msg)

    await download_file_http(
        url, output_file, timeout_seconds=timeout_seconds, chunk_size=chunk_size
    )

    return output_file



"""Module of ionex manipulation tools"""

from __future__ import annotations

from typing import NamedTuple

import numpy as np
from astropy.time import Time
from numpy.typing import NDArray

from spinifex.geometry.get_ipp import IPP
from spinifex.ionospheric.index_tools import (
    compute_index_and_weights,
    get_indices_axis,
)
from spinifex.ionospheric.ionex_download import (
    IonexOptions,
    _download_ionex,
)
from spinifex.ionospheric.ionex_parser import (
    IonexData,
    read_ionex,
    unique_days_from_ionex_files,
)
from spinifex.ionospheric.tec_data import ElectronDensity
from spinifex.times import get_indexlist_unique_days


class GroupedIPPs(NamedTuple):
    """Grouped IPPs"""

    ipps: list[IPP]
    indices: list[NDArray[np.int32]]


def interpolate_ionex(
    ionex: IonexData,
    lons: NDArray[np.float64],
    lats: NDArray[np.float64],
    times: Time,
    apply_earth_rotation: float = 1,
    get_rms: bool = False,
) -> NDArray[np.float64]:
    """Interpolate ionex data to a given lon/lat/height grid.
    lons, lats, times all should have the same length

    apply_earth_rotation:
    This is assuming that the TEC maps move according to the rotation
    Earth (following method 3 of interpolation described in the IONEX
    document). Experiments with high time resolution ROB data show that
    this is not really the case, resulting in strange wavelike structures
    when applying this smart interpolation.
    TODO: implement smoothing filters for interpolation

    Parameters
    ----------
    ionex : IonexData
        ionex object containing the information of the ionex file
    lons : NDArray
        longitudes (deg) of all points to interpolate to
    lats : NDArray
        lattitudes (deg) of all points to interpolate to
    times : Time
        times of all points to interpolate to
    apply_earth_rotation : float, optional
        specify (with a number between 0 and 1) how much of the earth rotation
        is taken in to account in the interpolation step., by default 1
    get_rms : bool, optional
        use rms values instead of tec values

    Returns
    -------
    NDArray
        array with interpolated tec values
    """
    value_array = ionex.rms if get_rms else ionex.tec

    timeindex = compute_index_and_weights(ionex.times.mjd, times.mjd)
    latindex = compute_index_and_weights(ionex.lats, lats)
    # take into account earth rotation
    if apply_earth_rotation > 0:
        rot1 = (
            (times.mjd - ionex.times.mjd[timeindex.idx1]) * 360.0
        ) * apply_earth_rotation
        rot2 = (
            (times.mjd - ionex.times.mjd[timeindex.idx2]) * 360.0
        ) * apply_earth_rotation
        lonindex1 = get_indices_axis(lons + rot1, ionex.lons, wrap_unit=360)
        lonindex2 = get_indices_axis(lons + rot2, ionex.lons, wrap_unit=360)
    else:
        lonindex1 = get_indices_axis(lons, ionex.lons, wrap_unit=360)
        lonindex2 = lonindex1
    tecdata = (
        value_array[timeindex.idx1, lonindex1.idx1, latindex.idx1]
        * lonindex1.w1
        * timeindex.w1
        * latindex.w1
    )
    tecdata += (
        value_array[timeindex.idx1, lonindex1.idx2, latindex.idx1]
        * lonindex1.w2
        * timeindex.w1
        * latindex.w1
    )

    tecdata += (
        value_array[timeindex.idx2, lonindex2.idx1, latindex.idx1]
        * lonindex2.w1
        * timeindex.w2
        * latindex.w1
    )
    tecdata += (
        value_array[timeindex.idx2, lonindex2.idx2, latindex.idx1]
        * lonindex2.w2
        * timeindex.w2
        * latindex.w1
    )

    tecdata += (
        value_array[timeindex.idx1, lonindex1.idx1, latindex.idx2]
        * lonindex1.w1
        * timeindex.w1
        * latindex.w2
    )
    tecdata += (
        value_array[timeindex.idx1, lonindex1.idx2, latindex.idx2]
        * lonindex1.w2
        * timeindex.w1
        * latindex.w2
    )

    tecdata += (
        value_array[timeindex.idx2, lonindex2.idx1, latindex.idx2]
        * lonindex2.w1
        * timeindex.w2
        * latindex.w2
    )
    tecdata += (
        value_array[timeindex.idx2, lonindex2.idx2, latindex.idx2]
        * lonindex2.w2
        * timeindex.w2
        * latindex.w2
    )
    return tecdata


def get_density_ionex(
    ipp: IPP, ionex_options: IonexOptions | None = None
) -> NDArray[np.float64]:
    """read ionex files and interpolate values to ipp locations/times

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints
    ionex_options : IonexOptions, optional
        optional arguments for the ionospheric model, by default None

    Returns
    -------
    NDArray
        array with tec values for every entry in ipp

    Raises
    ------
    FileNotFoundError
        if ionex file cannot be downloaded
    """
    # TODO: apply_earth_rotation as option
    sorted_ionex_paths = _download_ionex(times=ipp.times, options=ionex_options)

    unique_days = unique_days_from_ionex_files(sorted_ionex_paths)
    if not unique_days.shape:
        ionex = read_ionex(sorted_ionex_paths[0])
        return interpolate_ionex(ionex, ipp.loc.lon.deg, ipp.loc.lat.deg, ipp.times)
    group_indices = get_indexlist_unique_days(unique_days, ipp.times)
    tec = np.zeros(ipp.loc.shape, dtype=float)
    electron_density_error = np.full(ipp.loc.shape, np.nan)
    for indices, ionex_file in zip(group_indices, sorted_ionex_paths):
        if not ionex_file.exists():
            msg = f"Ionex file {ionex_file} not found!"
            raise FileNotFoundError(msg)
        u_loc = ipp.loc[indices]
        u_times = ipp.times[indices]
        ionex = read_ionex(ionex_file, options=ionex_options)
        tec[indices] = interpolate_ionex(ionex, u_loc.lon.deg, u_loc.lat.deg, u_times)
        electron_density_error[indices] = interpolate_ionex(
            ionex, u_loc.lon.deg, u_loc.lat.deg, u_times, get_rms=True
        )

    return ElectronDensity(
        electron_density=tec, electron_density_error=electron_density_error
    )



"""Testing of ionospheric indexing tools"""

from __future__ import annotations

import numpy as np
import pytest
from spinifex.exceptions import ArrayShapeError
from spinifex.ionospheric.index_tools import (
    compute_index_and_weights,
    get_indices,
    get_indices_axis,
    get_interpol,
    get_sorted_indices,
)


def test_get_indices():
    """test that indexing tools behave as expected"""
    idx = get_indices(5, np.arange(-10, 10))
    assert idx.idx1 == 15
    assert idx.idx2 == 16
    assert np.isclose(idx.w1, 1.0, 0.001)
    idx = get_indices(355, np.arange(0, 360, 10), wrap_unit=360.0)
    assert idx.idx1 == 0
    assert idx.idx2 == 35
    assert np.isclose(idx.w1, 0.5, 0.001)

    idx = get_indices_axis(np.array(355), np.arange(0, 360, 10), wrap_unit=360.0)
    assert idx.idx1 == 35
    assert idx.idx2 == 0
    assert np.isclose(idx.w1, 0.5, 0.001)
    idx = get_indices_axis(
        np.array([355, 700, 2]), np.arange(0, 360, 10), wrap_unit=360
    )
    assert np.array_equal(idx.idx1, np.array([35, 33, 0]))
    idx = get_indices_axis(
        np.arange(360).reshape(10, 36), np.arange(0, 360, 10), wrap_unit=360
    )
    assert idx.idx1.shape == (10, 36)


def test_get_sorted_indices():
    """test that sorted_indices behaves as expected"""
    with pytest.raises(ArrayShapeError):
        idx = get_sorted_indices(
            lon=350,
            lat=20,
            avail_lat=np.linspace(0, 180, 10),
            avail_lon=np.linspace(0, 360, 36),
            wrap_unit=360,
        )

    idx = get_sorted_indices(
        lon=350,
        lat=20,
        avail_lat=np.linspace(0, 180, 10),
        avail_lon=np.linspace(0, 360, 10),
        wrap_unit=360,
    )
    assert np.array_equal(idx.indices, np.array([0, 1, 2, 3, 7, 8, 6, 9, 5, 4]))


def test_compute_index_and_weights():
    """test that indices and weights are correctly computed"""
    idx = compute_index_and_weights(
        maparray=np.linspace(0, 180, 10), mapvalues=np.linspace(5, 175, 4)
    )
    assert np.array_equal(idx.idx1, np.array([0, 3, 5, 8]))
    assert idx.w2[0] == 0.25
    idx = compute_index_and_weights(maparray=np.linspace(0, 180, 10), mapvalues=55)
    assert idx.idx1.shape == (1,)


def test_get_interpol():
    """tests if interpolation behaves as expected"""
    value = get_interpol(
        data=np.linspace(0, 180, 10), dist=np.abs(np.arange(-50, 50, 10))
    )
    assert value == 100
    value = get_interpol(
        data=np.linspace(0, 180, 10), dist=np.array([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    )
    assert value == 50



"""Module to parse the UPC-Ionsat tomion data format"""

from __future__ import annotations

import asyncio
import gzip
from datetime import datetime
from pathlib import Path
from typing import Any, NamedTuple

import astropy.units as u
import numpy as np
from astropy.table import Table
from astropy.time import Time
from numpy.typing import NDArray
from pydantic import Field

from spinifex.asyncio_wrapper import sync_wrapper
from spinifex.download import download_or_copy_url
from spinifex.exceptions import TomionError
from spinifex.geometry import IPP
from spinifex.ionospheric.index_tools import (
    compute_index_and_weights,
    get_interpol,
    get_sorted_indices,
)
from spinifex.logger import logger
from spinifex.options import Options
from spinifex.times import get_indexlist_unique_days, get_unique_days

TOMOION_FORMAT_DICT: dict[str, Any] = {
    "mjd": float,
    "index": int,
    "value": float,
    "stddev": float,
    "type": str,
    "number_of_observations": int,
    "height": float,
    "ra": float,
    "dec": float,
    "i": int,
    "j": int,
    "k": int,
    "label": str,
    "longitude": float,
    "lst": float,
    "year": int,
    "doy": int,
    "month": int,
    "dom": int,
}
MAX_INTERPOL_POINTS: int = 4  # number of points used for lon/lat interpolation
TOMION_HEIGHTS: u.Quantity = [450, 1150] * u.km


class TomionOptions(Options):
    """Options for tomion model"""

    output_directory: Path | None = Field(
        None, description="Output directory for tomion files"
    )


class TomionData(NamedTuple):
    """Object containing all necessary information from Tomion data"""

    lons: NDArray[np.float64]
    """array with longitude values (degrees)"""
    lats: NDArray[np.float64]
    """array with latitude values (degrees)"""
    available_times: Time
    """array with unique times"""
    times: Time
    """times"""
    h: NDArray[np.float64]
    """heights (km)"""
    tec: NDArray[np.float64]
    """array with voxel tecvalues(TECU)"""
    h_idx: NDArray[np.int128]
    """array with index of height"""


def _read_tomion(fname: Path) -> TomionData:
    """reads a tomion format file and returns a TomionData object

    Parameters
    ----------
    fname : Path
        filename

    Returns
    -------
    TomionData
        object with data and axes of the data
    """
    try:
        tomion_data = Table.read(
            fname,
            format="ascii",
            names=list(TOMOION_FORMAT_DICT.keys()),
        )
    except Exception as e:
        msg = f"Could not read tomion file {fname}"
        raise TomionError(msg) from e

    time_column = tomion_data["mjd"].value
    available_times = Time(np.unique(time_column), format="mjd")
    times = Time(time_column, format="mjd")
    height = tomion_data["height"].value
    layer_height = np.ptp(height)
    layer_index = tomion_data["k"].value
    # TODO: What are these numbers?
    electron_density = (10.0 / 1.05) * layer_height * tomion_data["value"].value
    available_lat = tomion_data["dec"].value
    available_lon = tomion_data["longitude"].value
    return TomionData(
        lons=available_lon,
        lats=available_lat,
        available_times=available_times,
        times=times,
        h=height,
        h_idx=layer_index,
        tec=electron_density,
    )


async def get_tomion_paths_coro(
    unique_days: Time, tomion_options: TomionOptions | None = None
) -> list[Path]:
    """download tomion data for all unique days

    Parameters
    ----------
    unique_days : Time
        days for which to download data
    tomion_options : TomionOptions | None, optional
        options for the tomion model, by default None

    Returns
    -------
    list[Any]
        list of paths to the files
    """
    if unique_days.isscalar:
        unique_days = Time([unique_days])
    tomion_paths_coros = []
    for day in unique_days:
        url, nefull_name = _tomion_format(day)
        msg = f"downloading {url} to {nefull_name}"
        logger.info(msg)
        tomion_paths_coros.append(
            _download_tomion_file(
                url=url, nefull_name=nefull_name, tomion_options=tomion_options
            )
        )
    return await asyncio.gather(*tomion_paths_coros)


get_tomion_paths = sync_wrapper(get_tomion_paths_coro)


def _tomion_format(time: Time) -> tuple[str, str]:
    """helper function to get the url of the tomion data

    Parameters
    ----------
    time : Time
        day for which to get the url

    Returns
    -------
    tuple[Any, Any]
        url and the name how the data will be stored on disc
    """
    assert time.isscalar, "Only one time is supported"
    url_stem = "http://cabrera.upc.es/upc_ionex_GPSonly-RINEXv3/"
    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    # YYYY/DDD_YYMMDD.15min/.bias_dens
    nefull_name = f"NeFull.{dtime.year:04d}{doy:03d}"
    yy = f"{dtime.year:02d}"[-2:]
    file_name = f"bias_dens.0001.{dtime.year:04d}{doy:03d}.gz"
    directory_name = f"{dtime.year:04d}/{doy:03d}_{yy}{dtime.month:02d}{dtime.day:02d}.15min/.bias_dens"
    return f"{url_stem}/{directory_name}/{file_name}", nefull_name


async def _download_tomion_file(
    url: str,
    nefull_name: str,
    tomion_options: TomionOptions | None = None,
    timeout_seconds: int = 30,
    chunk_size: int = 1000,
) -> Path:
    """download and convert tomion file to readable nefull text file or if the file already exists
    just return a pointer to the file

    Parameters
    ----------
    url : str
        url of the file to download
    nefull_name : str
        name of the extracted nefull file
    tomion_options : TomionOptions | None, optional
        options for the ionospheric model, by default None
    timeout_seconds : int, optional
        time out for downloading, by default 30
    chunk_size : int, optional
        chunksize for downloading, by default 1000

    Returns
    -------
    Path
        pointer to the nefull file

    Raises
    ------
    IonexError
        error if the download times out
    """
    # TODO: Consider merging with download tools
    if tomion_options is None or tomion_options.output_directory is None:
        output_directory = Path.cwd() / "tomion_files"
    else:
        output_directory = tomion_options.output_directory

    output_file = output_directory / nefull_name

    if output_file.exists():
        msg = f"File {output_file} already exists. Skipping download."
        logger.info(msg)
        return output_file

    tomion_file = await download_or_copy_url(
        url=url,
        output_directory=output_directory,
        chunk_size=chunk_size,
        timeout_seconds=timeout_seconds,
    )

    output_file = await _extract_nefull(
        tomion_file=tomion_file, output_file=output_file
    )
    assert output_file.exists(), f"File {output_file} not found!"
    return output_file


async def _extract_nefull(
    tomion_file: Path, output_file: Path, search_term: str = "NeFull"
) -> Path:
    """helper function to get the relevant information from the large tomion file.
    Deletes the tomion file if the extraction was successful

    Parameters
    ----------
    tomion_file : Path
        pointer to the tomion file
    output_file : Path
        name of the nefull file
    search_term : str, optional
        the indicator for useful data in the tomion file, by default "NeFull"

    Returns
    -------
    Path
        pointer to the nefull file
    """
    with gzip.open(tomion_file, "rt") as f_in, output_file.open("w") as f_out:
        for line in f_in:
            if search_term in line:
                f_out.write(line)
    # Verify data was written correctly
    if Path.exists(output_file) and Path(output_file).stat().st_size > 0:
        msg = f"Extraction successful! Output saved in {output_file}"
        logger.info(msg)
        tomion_file.unlink()
        return output_file

    msg = f"Could not convert {tomion_file} to {output_file}"
    raise TomionError(msg)


def interpolate_tomion(
    tomion: TomionData,
    lons: NDArray[np.float64],
    lats: NDArray[np.float64],
    times: Time,
) -> NDArray[np.float64]:
    """Interpolate tomion data to the requested lons/lats/times

    Parameters
    ----------
    tomion : TomionData
        data object
    lons : NDArray[np.float64]
        array of longitudes at the two TOMION_HEIGHTS, shape (2,)
    lats : NDArray[np.float64]
        array of latitudes at the two TOMION_HEIGHTS, shape (2,)
    times : Time
        time

    Returns
    -------
    NDArray[np.float64]
        electron density values at two TOMION_HEIGHTS, shape (2,)
    """
    # TODO: implement apply_earth_rotation
    # TODO: implement this function directly for an array of times
    timeindex = compute_index_and_weights(tomion.available_times.mjd, times.mjd)
    time1 = tomion.available_times.mjd[timeindex.idx1]
    time2 = tomion.available_times.mjd[timeindex.idx2]
    timeselect1 = tomion.times.mjd == time1
    timeselect2 = tomion.times.mjd == time2
    timeselect = [timeselect1, timeselect2]
    # get data for two layers for two times
    layers_lo = [tomion.h_idx[timeselect1] == 1, tomion.h_idx[timeselect2] == 1]
    layers_hi = [tomion.h_idx[timeselect1] == 2, tomion.h_idx[timeselect2] == 2]
    # get lon,lat idx for these
    tec_lo = []
    tec_hi = []
    tec = np.zeros((2,), dtype=float)
    for lo, tms in zip(layers_lo, timeselect):
        isorted_low = get_sorted_indices(
            lon=lons[0],
            lat=lats[0],
            avail_lon=tomion.lons[tms][lo],
            avail_lat=tomion.lats[tms][lo],
        )
        tec_lo.append(
            get_interpol(
                tomion.tec[tms][lo][isorted_low.indices[:MAX_INTERPOL_POINTS]],
                isorted_low.distance[:MAX_INTERPOL_POINTS],
            )
        )
    for hi, tms in zip(layers_hi, timeselect):
        isorted_hi = get_sorted_indices(
            lon=lons[1],
            lat=lats[1],
            avail_lon=tomion.lons[tms][hi],
            avail_lat=tomion.lats[tms][hi],
        )
        tec_hi.append(
            get_interpol(
                tomion.tec[tms][hi][isorted_hi.indices[:MAX_INTERPOL_POINTS]],
                isorted_hi.distance[:MAX_INTERPOL_POINTS],
            )
        )

    tec[0] = tec_lo[0] * timeindex.w1[0] + tec_lo[1] * timeindex.w2[0]
    tec[1] = tec_hi[0] * timeindex.w1[0] + tec_hi[1] * timeindex.w2[0]

    return tec


def get_density_dual_layer(
    ipp: IPP, tomion_options: TomionOptions | None = None
) -> NDArray[np.float64]:
    """extracts electron densities for the two TOMION_HEIGHTS for all times in ipp. The returned array
    will have zeros every where apart from the two altitudes closest to TOMION_HEIGHTS

    Parameters
    ----------
    ipp : IPP
        input piercepoint locations
    tomion_options : TomionOptions | None, optional
        optional ionospheric model options, by default None

    Returns
    -------
    NDArray[np.float64]
        array with electron densities shape (ipp.loc.shape). The array will only have values at the
        altitudes closest to the TOMION_HEIGHTS and zeros everywhere else.

    Raises
    ------
    FileNotFoundError
        error if the tomion files are not available locally nor online
    """
    # TODO: no need to go through all the burden, just make sure that the ipps are correct for this model?
    h_index = [
        np.argmin(np.abs(ipp.loc[0].height.to(u.km).value - h.to(u.km).value))
        for h in TOMION_HEIGHTS
    ]
    selected_ipp = ipp.loc[:, h_index]
    tec = np.zeros(ipp.loc.shape, dtype=float)
    unique_days = get_unique_days(times=ipp.times)
    sorted_tomion_paths = get_tomion_paths(
        unique_days=unique_days,
        tomion_options=tomion_options,
    )

    group_indices = get_indexlist_unique_days(unique_days, ipp.times)
    for indices, tomion_file in zip(group_indices, sorted_tomion_paths):
        if not tomion_file.exists():
            msg = f"Tomion file {tomion_file} not found!"
            raise FileNotFoundError(msg)
        u_loc = selected_ipp[indices]
        u_times = ipp.times[indices]
        tomion = _read_tomion(tomion_file)
        for idxi, ippi in enumerate(np.arange(tec.shape[0])[indices]):
            tec[ippi, h_index] = interpolate_tomion(
                tomion, u_loc[idxi].lon.deg, u_loc[idxi].lat.deg, u_times[idxi]
            )

    return tec



"""Module to calculate electron densities"""

from __future__ import annotations

from typing import Any, NamedTuple

import astropy.units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time
from numpy.typing import NDArray

from spinifex.geometry import IPP, get_ipp_from_altaz, get_ipp_from_skycoord
from spinifex.get_rm import DEFAULT_IONO_HEIGHT
from spinifex.ionospheric import ModelDensityFunction
from spinifex.ionospheric.models import (
    O,
    parse_iono_kwargs,
    parse_iono_model,
)
from spinifex.logger import logger


class DTEC(NamedTuple):
    """object with all electron densities"""

    times: Time
    """time axis"""
    electron_density: NDArray[Any]
    """electron content"""
    airmass: NDArray[Any]
    """conversion from vertical to slant TEC"""
    height: NDArray[Any]
    """array of altitudes (km)"""
    loc: EarthLocation
    """observer location"""


def _get_dtec(
    ipp: IPP,
    iono_model: ModelDensityFunction[O],
    iono_options: O | None = None,
) -> DTEC:
    """Get the electron densities for a given set of ionospheric piercepoints

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    iono_options : OptionType, optional
        options for the ionospheric model, by default None

    Returns
    -------
    DTEC
        electron densities object
    """
    logger.info("Calculating electron density")
    density_profile = iono_model(ipp=ipp, options=iono_options)
    return DTEC(
        times=ipp.times,
        electron_density=density_profile.electron_density,
        airmass=ipp.airmass,
        height=ipp.loc.height.to(u.km).value,
        loc=ipp.station_loc,
    )


def _get_dtec_from_altaz(
    loc: EarthLocation,
    altaz: AltAz,
    iono_model: ModelDensityFunction[O],
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_options: O | None = None,
) -> DTEC:
    """get electron densities for user defined altaz coordinates

    Parameters
    ----------
    loc : EarthLocation
        observer location
    altaz : AltAz
        altaz coordinates
    height_array : u.Quantity, optional
        altitudes, by default default_height
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    iono_options : dict
        keyword arguments for the ionospheric model


    Returns
    -------
    DTEC
        electron density object
    """
    ipp = get_ipp_from_altaz(loc=loc, altaz=altaz, height_array=height_array)
    return _get_dtec(
        ipp=ipp,
        iono_model=iono_model,
        iono_options=iono_options,
    )


def get_dtec_from_altaz(
    loc: EarthLocation,
    altaz: AltAz,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    **iono_kwargs: Any,
) -> DTEC:
    """get rotation measures for user defined altaz coordinates

    Parameters
    ----------
    loc : EarthLocation
        observer location
    altaz : AltAz
        altaz coordinates
    height_array : u.Quantity, optional
        altitudes, by default default_height
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    iono_options : dict
        keyword arguments for the ionospheric model


    Returns
    -------
    DTEC
        electron density object
    """
    iono_model = parse_iono_model(iono_model_name)
    iono_options = parse_iono_kwargs(iono_model=iono_model, **iono_kwargs)
    return _get_dtec_from_altaz(
        loc=loc,
        altaz=altaz,
        height_array=height_array,
        iono_model=iono_model,
        iono_options=iono_options,
    )


def _get_dtec_from_skycoord(
    loc: EarthLocation,
    times: Time,
    source: SkyCoord,
    iono_model: ModelDensityFunction[O],
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_options: O | None = None,
) -> DTEC:
    """get electron densities for user defined times and source coordinate

    Parameters
    ----------
    loc : EarthLocation
        observer location
    times : Time
        times
    source : SkyCoord
        coordinates of the source
    height_array : NDArray, optional
        altitudes, by default default_height
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    iono_kwargs : dict
        keyword arguments for the ionospheric model


    Returns
    -------
    DTEC
        relectron density object
    """

    ipp = get_ipp_from_skycoord(
        loc=loc, times=times, source=source, height_array=height_array
    )
    return _get_dtec(
        ipp=ipp,
        iono_model=iono_model,
        iono_options=iono_options,
    )


def get_dtec_from_skycoord(
    loc: EarthLocation,
    times: Time,
    source: SkyCoord,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    # iono_model: ModelDensityFunction[IonoOptions] = ionospheric_models.ionex,
    **iono_kwargs: Any,
) -> DTEC:
    """get electron densities for user defined times and source coordinate

    Parameters
    ----------
    loc : EarthLocation
        observer location
    times : Time
        times
    source : SkyCoord
        coordinates of the source
    height_array : NDArray, optional
        altitudes, by default default_height
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    iono_kwargs : dict
        keyword arguments for the ionospheric model


    Returns
    -------
    DTEC
        relectron density object
    """
    iono_model = parse_iono_model(iono_model_name)
    iono_options = parse_iono_kwargs(iono_model=iono_model, **iono_kwargs)
    return _get_dtec_from_skycoord(
        loc=loc,
        times=times,
        source=source,
        height_array=height_array,
        iono_model=iono_model,
        iono_options=iono_options,
    )



"""data objects for ionospheric data"""

from __future__ import annotations

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray


class ElectronDensity(NamedTuple):
    """object containing interpolated electron density values and their estimated uncertainty"""

    electron_density: NDArray[np.float64]
    """electron density in TECU"""
    electron_density_error: NDArray[np.float64]
    """uncertainty in TECU"""



"""Module for getting ionospheric ionospheric_models"""

from __future__ import annotations

from spinifex.ionospheric.models import (
    ModelDensityFunction,
    get_density_ionex_iri,
    get_density_ionex_single_layer,
    ionospheric_models,
)

__all__ = [
    "ModelDensityFunction",
    "get_density_ionex_iri",
    "get_density_ionex_single_layer",
    "get_density_tomion",
    "ionospheric_models",
]



"""Testing of ionospheric"""

from __future__ import annotations

from importlib import resources
from typing import Any

from astropy.utils import iers
from spinifex.ionospheric.models import parse_iono_kwargs

iers.conf.auto_download = False

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from spinifex.geometry.get_ipp import IPP, get_ipp_from_skycoord
from spinifex.ionospheric import ionospheric_models
from spinifex.ionospheric.ionex_parser import (
    read_ionex,
    unique_days_from_ionex,
    unique_days_from_ionex_files,
)


@pytest.fixture
def ipp() -> IPP:
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2020-01-08T01:00:00") + np.arange(0, 10) * 15 * u.min
    return get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )


@pytest.fixture
def ipp2() -> IPP:
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2020-01-08T01:00:00") + np.arange(0, 10) * 3.5 * u.hr
    return get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )


def test_get_ionosphere(ipp):
    """Test that get_ionosphere does not crash"""
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        testdata = datapath / "codg0080.20i.Z"

    ionex = read_ionex(testdata)
    assert ionex.tec.shape == (25, 73, 71)

    tec = ionospheric_models.ionex(ipp)
    assert tec.electron_density.shape == ipp.loc.shape
    tec = ionospheric_models.ionex_iri(ipp)
    assert tec.electron_density.shape == ipp.loc.shape


def test_read_zcompressed():
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        testdata = datapath / "casg0010.99i.Z"
    ionex = read_ionex(testdata)
    assert ionex.tec.shape == (12, 73, 71)


def test_ionosphere_ionex(ipp):
    iono_kwargs: dict[str, Any] = {}
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        iono_kwargs["output_directory"] = datapath
        iono_kwargs["prefix"] = "esa"
        iono_kwargs["server"] = "cddis"
        options = parse_iono_kwargs(ionospheric_models.ionex, **iono_kwargs)
        tec = ionospheric_models.ionex(ipp, options=options)
        assert tec.electron_density.shape == ipp.loc.shape
        assert tec.electron_density_error.shape[0] == ipp.loc.shape[0]

        # Test bad arguments
        with pytest.raises(TypeError):
            options = parse_iono_kwargs(ionospheric_models.ionex, bad_arg="bad")


def test_ionosphere_ionex_multiple_days(ipp2):
    iono_kwargs: dict[str, Any] = {}
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        iono_kwargs["output_directory"] = datapath
        iono_kwargs["prefix"] = "esa"
        iono_kwargs["server"] = "cddis"
        options = parse_iono_kwargs(ionospheric_models.ionex, **iono_kwargs)
        tec = ionospheric_models.ionex(ipp2, options=options)
        assert tec.electron_density.shape == ipp2.loc.shape
        assert tec.electron_density_error.shape[0] == ipp2.loc.shape[0]


def test_unique_days():
    with resources.as_file(resources.files("spinifex.data.tests")) as datapath:
        testdata = datapath / "codg0080.20i.Z"
        other_data = datapath / "casg0010.99i.Z"

    unique_days = unique_days_from_ionex_files(testdata)

    assert unique_days.value.shape == (1,)
    assert unique_days.value[0] == 58856.0

    ionex = read_ionex(testdata)
    unique_days = unique_days_from_ionex(ionex)
    assert unique_days.value.shape == (1,)
    assert unique_days.value[0] == 58856.0

    unique_days = unique_days_from_ionex_files([testdata, testdata])
    assert unique_days.value.shape == (1,)
    assert unique_days.value[0] == 58856.0

    ionex = read_ionex(testdata)
    unique_days = unique_days_from_ionex([ionex, ionex])
    assert unique_days.value.shape == (1,)
    assert unique_days.value[0] == 58856.0

    unique_days = unique_days_from_ionex_files([testdata, other_data])
    assert unique_days.value.shape == (2,)
    assert unique_days.value[0] == 51179.0
    assert unique_days.value[1] == 58856.0



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, NamedTuple, Protocol, TypeVar, cast


class FooOptions(NamedTuple):
    foo: str


class BarOptions(NamedTuple):
    bar: str


O = TypeVar("O", FooOptions, BarOptions)  # noqa: E741
O_contra = TypeVar("O_contra", FooOptions, BarOptions, contravariant=True)


class FooBarFunction(Protocol, Generic[O_contra]):
    def __call__(self, options: O_contra | None) -> str: ...


@dataclass
class FooBarFunctions:
    foo: FooBarFunction[FooOptions]
    bar: FooBarFunction[BarOptions]


def foo_func(options: FooOptions | None = None) -> str:
    if options is None:
        return "foo"
    return options.foo


def bar_func(options: BarOptions | None = None) -> str:
    if options is None:
        return "bar"
    return options.bar


foo_bar_functions = FooBarFunctions(
    foo=foo_func,
    bar=bar_func,
)


def parse_options(foo_bar_func: FooBarFunction[O], **kwargs: Any) -> O:
    if foo_bar_func == foo_bar_functions.foo:
        return cast(O, FooOptions(**kwargs))
    if foo_bar_func == foo_bar_functions.bar:
        return cast(O, BarOptions(**kwargs))

    msg = "Unknown function type"
    raise TypeError(msg)


def parse_foo_bar_func(foo_bar_name: str) -> FooBarFunction[O]:
    try:
        return getattr(foo_bar_functions, foo_bar_name)  # type: ignore[no-any-return]
    except AttributeError as e:
        msg = f"Unknown function {foo_bar_name}. Supported functions are {list(foo_bar_functions.__annotations__.keys())}"
        raise TypeError(msg) from e


def baz(
    foo_bar_name: str = "foo",
    **kwargs: Any,
) -> str:
    foo_bar_func = parse_foo_bar_func(foo_bar_name)
    options = parse_options(foo_bar_func, **kwargs)
    return _baz(
        foo_bar_func=foo_bar_func,
        options=options,
    )


def _baz(
    foo_bar_func: FooBarFunction[O],
    options: O | None = None,
) -> str:
    return foo_bar_func(options=options)
